import unittest
import mock
import sys
import os
sys.path.insert(0, os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.abspath(os.path.dirname(__file__))),
    "lazy_record"))
import typecasts
import datetime


class TestTypes(unittest.TestCase):

    def test_date_casts_date_to_date(self):
        self.assertEqual(datetime.date(2016, 1, 1),
            typecasts.date(datetime.date(2016, 1, 1)))

    def test_date_casts_datetime_to_datetime(self):
        self.assertEqual(datetime.datetime(2016, 1, 1),
            typecasts.datetime(datetime.datetime(2016, 1, 1)))


if __name__ == '__main__':
    unittest.main()



import unittest
import mock
import sys
import os
sys.path.insert(0, os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.abspath(os.path.dirname(__file__))),
    "lazy_record"))
from repo import Repo
import repo


class TunaCasserole(object):
    pass


@mock.patch("repo.Repo.db")
class TestRepo(unittest.TestCase):

    def test_gets_table_name(self, _):
        table_name = Repo.table_name(TunaCasserole)
        self.assertEqual("tuna_casseroles", table_name)

    @mock.patch("repo.sqlite3")
    def test_connects_database(self, sqlite3, db):
        repo.Repo.connect_db("my_db")
        sqlite3.connect.assert_called_with("my_db",
            detect_types=sqlite3.PARSE_DECLTYPES)
        self.assertEqual(repo.Repo.db, sqlite3.connect.return_value)

    def test_makes_query_for_all_records(self, db):
        Repo("tuna_casseroles").select("id", "created_at")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.id, tuna_casseroles.created_at "
            "from tuna_casseroles", [])

    def test_where_restricts_query(self, db):
        Repo("tuna_casseroles").where(my_attr=5).select("id", "created_at")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.id, tuna_casseroles.created_at "
            "from tuna_casseroles "
            "where tuna_casseroles.my_attr == ?", [5])

    def test_where_allows_arbitrary_restriction(self, db):
        Repo("tuna_casseroles").where([("my_attr > ?", 15)]).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.my_attr > ?", [15])

    def test_arbitrary_where_does_not_double_scope_column(self, db):
        Repo("tuna_casseroles").where([("tuna_casseroles.my_attr > ?",
                                        15)]).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.my_attr > ?", [15])

    def test_arbitrary_where_combines_with_regular_where(self, db):
        Repo("tuna_casseroles").where([("my_attr > ?",
                                        15)], name="foo").select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.name == ? "
            "and tuna_casseroles.my_attr > ?", ["foo", 15])

    def test_arbitrary_where_can_accept_no_args(self, db):
        Repo("tuna_casseroles").where([("my_attr == 11",)]).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.my_attr == 11", [])

    def test_arbitrary_where_can_accept_many_args(self, db):
        Repo("tuna_casseroles").where([("my_attr == ? or my_attr == ?", 5, 11)],
            ).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.my_attr == ? "
            "or tuna_casseroles.my_attr == ?", [5, 11])

    def test_accepts_multiple_arbitrary_wheres(self, db):
        Repo("tuna_casseroles").where([("my_attr == ?", 5),
                                       ("my_attr == ?", 11)],
                                      ).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.my_attr == ? "
            "and tuna_casseroles.my_attr == ?", [5, 11])

    def test_queries_through_tables(self, db):
        Repo("tuna_casseroles").where(
            my_relations=dict(my_attr=5)).select("id", "created_at")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.id, tuna_casseroles.created_at "
            "from tuna_casseroles "
            "where my_relations.my_attr == ?",
            [5])

    def test_joins_tables(self, db):
        Repo("tuna_casseroles").inner_join(
            {'table': "my_relations", 'on': ["tuna_casserole_id", "id"]}
            ).select("id", "created_at")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.id, tuna_casseroles.created_at "
            "from tuna_casseroles "
            "inner join my_relations on "
            "my_relations.tuna_casserole_id == tuna_casseroles.id",
            [])

    def test_joins_multiple_tables(self, db):
        # Repo("tuna_casseroles").inner_join("my_relations",
        #     on=["tuna_casserole_id", "id"]).inner_join("my_others",
        #     on=["id", "my_other_id"]).select("id", "created_at")
        Repo("tuna_casseroles").inner_join(
            {'table': "my_relations", 'on': ["tuna_casserole_id", "id"]},
            {'table': "my_others", 'on': ["id", "my_other_id"]}
            ).select("id", "created_at")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.id, tuna_casseroles.created_at "
            "from tuna_casseroles "
            "inner join my_relations on "
            "my_relations.tuna_casserole_id == tuna_casseroles.id "
            "inner join my_others on "
            "my_others.id == my_relations.my_other_id",
            [])

    def test_joins_tables_with_where(self, db):
        Repo("tuna_casseroles").inner_join(
            {'table': "my_relations", 'on': ["tuna_casserole_id", "id"]}
            ).where(my_relations=dict(my_attr=5)
            ).select("id", "created_at")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.id, tuna_casseroles.created_at "
            "from tuna_casseroles "
            "inner join my_relations on "
            "my_relations.tuna_casserole_id == tuna_casseroles.id "
            "where my_relations.my_attr == ?",
            [5])

    def test_inserts_records(self, db):
        Repo("tuna_casseroles").insert(my_attr=7)
        db.execute.assert_called_once_with(
            "insert into tuna_casseroles (my_attr) values (?)", [7])

    def test_returns_id_of_inserted_record(self, db):
        db.execute.return_value.lastrowid = 5
        self.assertEqual(Repo("tuna_casseroles").insert(my_attr=7), 5)

    @mock.patch("repo.Repo.db")
    def test_does_not_insert_records_with_where(self, db,  _):
        with self.assertRaises(repo.Invalid):
            repo.Repo("tuna_casseroles").where(x=5).insert(my_attr=7)
        self.assertEqual(db.execute.call_count, 0,
            "expected not to execute db command, but did.")

    def test_updates_records(self, db):
        Repo("tuna_casseroles").update(my_attr=7)
        db.execute.assert_called_once_with(
            "update tuna_casseroles set my_attr = ?", [7])

    def test_updates_with_where(self, db):
        Repo("tuna_casseroles").where(id=15).update(my_attr=7)
        db.execute.assert_called_once_with(
            "update tuna_casseroles set my_attr = ? "
            "where tuna_casseroles.id == ?", [7, 15])

    def test_deletes_records(self, db):
        Repo("tuna_casseroles").delete()
        db.execute.assert_called_once_with(
            "delete from tuna_casseroles", [])

    def test_deletes_records_with_where(self, db):
        Repo("tuna_casseroles").where(id=11).delete()
        db.execute.assert_called_once_with(
            "delete from tuna_casseroles where tuna_casseroles.id == ?", [11])

    def test_orders_records(self, db):
        Repo("tuna_casseroles").order_by(id="desc").select("id", "created_at")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.id, tuna_casseroles.created_at from "
            "tuna_casseroles order by id desc", [])

    def test_gets_count_of_records(self, db):
        Repo("tuna_casseroles").count()
        db.execute.assert_called_once_with(
            "select COUNT(*) from tuna_casseroles", [])

    def test_gets_count_of_records_with_other_query_elements(self, db):
        Repo("tuna_casseroles").where(id=11).count()
        db.execute.assert_called_once_with(
            "select COUNT(*) from tuna_casseroles "
            "where tuna_casseroles.id == ?", [11])

    def test_where_with_list_generates_in(self, db):
        Repo("tuna_casseroles").where(name=["foo", "bar", "baz"]).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.name IN (?, ?, ?)",
            ["foo", "bar", "baz"])

    def test_where_with_list_generates_in_with_multiple(self, db):
        Repo("tuna_casseroles").where(name=["foo", "bar", "baz"],
                                      id=(1,2)).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.name IN (?, ?, ?) "
            "and tuna_casseroles.id IN (?, ?)",
            ["foo", "bar", "baz", 1, 2])

    def test_group_generates_group_by_clause(self, db):
        Repo("tuna_casseroles").group_by("name").select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "GROUP BY name", [])

    def test_having_generates_having_clause(self, db):
        Repo("tuna_casseroles").group_by("name"
                              ).having([("sum(value) > ?", 11)]).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "GROUP BY name "
            "HAVING sum(value) > ?", [11])

    def test_where_and_having_puts_values_in_correct_order(self, db):
        Repo("tuna_casseroles").where(id=87
                              ).group_by("name"
                              ).having([("sum(value) > ?", 11)]).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.id == ? "
            "GROUP BY name "
            "HAVING sum(value) > ?", [87, 11])

    def test_limit_limits_number_of_returned_objects(self, db):
        Repo("tuna_casseroles").where(id=87
                              ).limit(10).select("*")
        db.execute.assert_called_once_with(
            "select tuna_casseroles.* from tuna_casseroles "
            "where tuna_casseroles.id == ? "
            "LIMIT ?", [87, 10]
        )

    def test_limit_errors_if_limit_is_zero(self, db):
        with self.assertRaises(repo.Invalid):
            Repo("tuna_casseroles").where(id=87
                                  ).limit(0).select("*")

if __name__ == '__main__':
    unittest.main()



import query
import repo
from lazy_record.errors import *
from inflector import Inflector, English

inflector = Inflector(English)

models = {}
associations = {}
foreign_keys = {}
scopes = {}

def model_from_name(parent_name):
    return models[inflector.classify(parent_name)]

def _verify_type_match(record, association):
    associated_model = model_from_name(association)
    if record is None:
        return
    if not isinstance(record, associated_model):
        raise AssociationTypeMismatch(
            "Expected record of type {expected}, got {actual}.".format(
                expected=associated_model.__name__,
                actual=record.__class__.__name__
            ))

def model_has_foreign_key_for_table(table, model):
    fk = foreign_keys_for(model).get(inflector.singularize(table), None)
    if fk is None:
        return True
    return fk in model.__attributes__

def foreign_keys_for(klass):
    if type(klass) == str:
        klass_name = klass
    else:
        klass_name = klass.__name__
    foreign_keys[klass_name] = foreign_keys.get(klass_name, {})
    return foreign_keys[klass_name]

def associations_for(klass):
    if type(klass) == str:
        klass_name = klass
    else:
        klass_name = klass.__name__
    associations[klass_name] = associations.get(klass_name, {})
    return associations[klass_name]

def scopes_for(klass):
    klass_name = klass.__name__
    scopes[klass_name] = scopes.get(klass_name, {})
    return scopes[klass_name]

class belongs_to(object):
    """
    Decorator to establish this model as the child in a one-to-many
    relationship.
    """
    def __init__(self, parent_name, foreign_key=None):
        """
        +parent_name+ is the parent model (e.g. if a post has many comments,
        the comment will have one post: post is the +parent_name+).
        +foreign_key+ is the foreign key used in the child (this) record to
        hold the id of the parent record. By default it is the parent's model
        name, snake-cased, with "_id" appended (e.g. Post -> "post_id").

        Creates a setter and getter property for the parent record.

        ex)

        >>> comment.post_id
        1
        >>> comment.post
        Post(id=1)
        >>> other_post
        Post(id=1)
        >>> comment.post = other_post
        >>> comment.post
        Post(id=2)
        >>> comment.post_id
        2
        """
        self.parent_name = parent_name
        self.foreign_key = foreign_key or inflector.foreignKey(parent_name)

    def __call__(self, klass):
        # Add the model to the registry of known models with associations
        models[klass.__name__] = klass
        # Set the foreign key in the model in case it needs to be looked up
        foreign_keys_for(klass)[self.parent_name] = self.foreign_key
        # Add the relationship to the association list
        associations_for(klass)[self.parent_name] = None

        # Getter method for the parent record (e.g. comment.post)
        # Is added to the class as a property
        def parent_record_getter(wrapped_obj):
            parent = model_from_name(self.parent_name)
            # Not using parent.find() because it raises if it cannot find
            q = query.Query(parent)
            return q.where(id=getattr(wrapped_obj, self.foreign_key)).first()

        # Setter method for updating the foreign key in this object
        def parent_record_setter(wrapped_obj, new_parent):
            if new_parent is not None:
                _verify_type_match(new_parent, self.parent_name)
                # We are setting a parent: grab it's id and use it
                setattr(wrapped_obj, self.foreign_key, new_parent.id)
            else:
                # Un-setting a parent, set the foreign key to None
                # Can't use new_parent.id since new_parent is None
                setattr(wrapped_obj, self.foreign_key, None)

        # Add setter and getter to class as properties
        setattr(klass, self.parent_name,
                property(parent_record_getter, parent_record_setter))
        # Add the foreign key to the attribute dict of the model
        # Doing so in such a way as to not mutate the dict, otherwise it can
        # override the value in lazy_record.Base (and thus all models)
        new_attributes = dict(klass.__attributes__)
        new_attributes[self.foreign_key] = int
        klass.__attributes__ = new_attributes
        return klass


# Currently exists only so that all models get registered
class has_many(object):
    """
    Decorator to establish this model as the parent in a one-to-many
    relationship or as one part of a many-to-many relationship
    """
    def __init__(self, child_name, scope=lambda query: query,
                 foreign_key=None, through=None):
        """
        +child_name+ is the child model (e.g. if a post has many comments:
        comments is the +child_name+). +foreign_key+ is the foreign key used in
        the child (this) record to hold the id of the parent record. By default
        it is the parent's model name, snake-cased, with "_id" appended
        (e.g. Post -> "post_id"). If this is a many-to-many relationship,
        +through+ is the joining table.

        Creates a getter property for child records and (if applicable the
        joining records).

        ex)

        >>> post.comments
        <lazy_record.Query [Comment(id=1)]>
        """
        self.child_name = child_name
        self.foreign_key = foreign_key
        self.through = through
        self.scope = scope

    def __call__(self, klass):
        self.klass = klass
        our_name = inflector.singularize(repo.Repo.table_name(klass))
        child_model_name = inflector.classify(self.child_name)
        scopes_for(klass)[self.child_name] = self.scope
        # If we are doing an implicit has_many using through, we should define it fully
        if self.through and self.through not in associations_for(klass):
            klass = has_many(self.through)(klass)
        if self.through and self.through not in associations_for(child_model_name):
            # Set up the association for the child
            # Assume a one-many tree unless already defined otherwise
            associations_for(child_model_name)[our_name] = \
                inflector.singularize(self.through)
        # if no foreign key was passed, we should calculate it now based on
        # the class name
        self.foreign_key = self.foreign_key or inflector.foreignKey(our_name)
        models[klass.__name__] = klass
        # Add the foreign key to the fk list
        if not self.through:
            foreign_keys_for(klass)[self.child_name] = self.foreign_key
            # Add the childs associations and foreign keys as if they had
            # a belongs_to
            foreign_keys_for(child_model_name)[our_name] = self.foreign_key
            associations_for(child_model_name)[our_name] = None
            
        # Add the relationship to the association list
        associations_for(klass)[self.child_name] = self.through

        # Add the child table (or joining table) to the classes dependents
        # so that if this record is destroyed, all related child records
        # (or joining records) are destroyed with it to prevent orphans
        if self.through:
            if self.through in foreign_keys_for(klass):
                klass.__dependents__ = klass.__dependents__ + [self.through]
        else:
            klass.__dependents__ = klass.__dependents__ + [self.child_name]
        if self.through:

            # Do the query with a join
            def child_records_method(wrapped_obj):
                child = model_from_name(self.child_name)
                # No guarentee that self.through is the last in the chain
                # It could be the other part of a many-to-many
                # Or it could be a through that is a couple of levels down
                # e.g. Category has many Posts through Threads
                #      (but chain is Category -> Forum -> Thread -> Post)
                result = query.Query(child, record=wrapped_obj).joins(
                            repo.Repo.table_name(wrapped_obj.__class__)).where(
                            **{repo.Repo.table_name(wrapped_obj.__class__):
                            {'id': wrapped_obj.id}})
                return self.scoping(result)
        else:
            # Don't do a join
            def child_records_method(wrapped_obj):
                child = model_from_name(self.child_name)
                q = query.Query(child, record=wrapped_obj)
                where_statement = {self.foreign_key: wrapped_obj.id}
                return self.scoping(q.where(**where_statement))

        setattr(klass, self.child_name, property(child_records_method))
        return klass

    def scoping(self, query):
        current = self.klass
        scopes = []
        while True:
            # Anything other than a has_many won't have an entry,
            # so we return the identity scope
            scopes.append(scopes_for(current).get(self.child_name,
                                                  lambda query: query))
            # Should give either a joining model or None
            # TODO handle case for has_one
            # currently, only has_many_supports scoping
            joiner = associations_for(current).get(self.child_name)
            # End of the line
            if joiner is None:
                break
            # Get the next in the list by looking at the joiner model
            current = model_from_name(joiner)
        for scope in scopes:
            query = scope(query)
        return query

class has_one(object):
    """
    Decorator to establish this model as the parent in a one-to-one
    relationship.
    """
    def __init__(self, child_name, foreign_key=None, through=None):
        self.child_name = child_name
        self.foreign_key = foreign_key
        self.through = through
        if str(self.through).endswith('s'):
            raise AssociationForbidden(
                "Cannot have one '{}' through many '{}'".format(
                    self.child_name,
                    self.through,
                ))

    def __call__(self, klass):
        our_name = inflector.singularize(repo.Repo.table_name(klass))
        child_model_name = inflector.classify(self.child_name)
        self.foreign_key = self.foreign_key or inflector.foreignKey(our_name)
        klass.__dependents__ = klass.__dependents__ + [self.child_name]
        # Add the relationship to the association list
        associations_for(klass)[self.child_name] = self.through
        # Add the foreign key to the fk list
        foreign_keys_for(klass)[self.child_name] = self.foreign_key
        models[klass.__name__] = klass
        if self.through and self.through not in associations_for(child_model_name):
            # Set up the association for the child
            # Assume a one-many tree unless already defined otherwise
            associations_for(child_model_name)[our_name] = self.through

        if self.through:

            def child_record_method(wrapped_obj):
                child = model_from_name(self.child_name)
                return query.Query(child, record=wrapped_obj).joins(
                          repo.Repo.table_name(wrapped_obj.__class__)).where(
                          **{repo.Repo.table_name(wrapped_obj.__class__):
                              {'id': wrapped_obj.id}}).first()

            def set_child_record_method(wrapped_obj, new_value):
                _verify_type_match(new_value, self.child_name)
                child = model_from_name(self.child_name)
                table = repo.Repo.table_name(wrapped_obj.__class__)
                q = query.Query(child, record=wrapped_obj).joins(table
                        ).where(**{table: {'id': wrapped_obj.id}})
                # Get the previous value
                old_value = q.first()
                # Recall that join_args will have either 0 or 2 or more,
                # never 1 element
                joiner = q.join_args[-2]
                # Find the intermediate record that will connect +new_value+
                # to wrapped_obj
                next_up = model_from_name(joiner['table'])
                next_r = query.Query(next_up, record=wrapped_obj).joins(
                             table).where(**{table: {'id': wrapped_obj.id}}
                             ).first()
                if not model_has_foreign_key_for_table(joiner['table'],
                                                       child):
                    # The intermediate record has the foreign key: set it
                    if new_value is None:
                        setattr(next_r,
                                joiner['on'][0],
                                None)
                    else:
                        setattr(next_r,
                                joiner['on'][0],
                                getattr(new_value, joiner['on'][1]))
                    wrapped_obj._related_records.append(next_r)
                else:
                    # Set the foreign key on the new value
                    if new_value is not None:
                        # Associate new value
                        setattr(new_value,
                                joiner['on'][1], # Foreign key
                                # Lookup the id/foreign_key of the record
                                getattr(next_r, joiner['on'][0]))
                        wrapped_obj._related_records.append(new_value)
                    # Disassociate the old value
                    if old_value is not None:
                        setattr(old_value,
                                joiner['on'][1], # Foreign key
                                None)
                        wrapped_obj._related_records.append(old_value)

        else:

            def child_record_method(wrapped_obj):
                child = model_from_name(self.child_name)
                q = query.Query(child, record=wrapped_obj)
                where_statement = {self.foreign_key: wrapped_obj.id}
                return q.where(**where_statement).first()

            def set_child_record_method(wrapped_obj, child):
                _verify_type_match(child, self.child_name)
                # We are setting a child: set its foreign key to our id
                if child is not None:
                    setattr(child, self.foreign_key, wrapped_obj.id)
                    wrapped_obj._related_records.append(child)
                # disassociate old record
                old_value = child_record_method(wrapped_obj)
                if old_value is not None:
                    setattr(old_value, self.foreign_key, None)
                    wrapped_obj._related_records.append(old_value)

        setattr(klass, self.child_name, property(child_record_method,
                                                 set_child_record_method))
        return klass



from repo import Repo
import sys
import os
import types
sys.path.insert(0, os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
from inflector import Inflector, English

inflector = Inflector(English)

def does_not_mutate(func):
    """Prevents methods from mutating the receiver"""
    def wrapper(self, *args, **kwargs):
        new = self.copy()
        return func(new, *args, **kwargs)
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper

class Query(object):
    """
    Generic Query object used for searching a database for records, and
    constructing records using the returned values from the database.
    """

    def __init__(self, model, record=None):
        """
        Instantiates a new Query. +model+ is the lazy_record model (i.e. class)
        that the query will be made about: the query's member objects will be
        members of +model+'s class. When +record+ is passed, it is used for
        managing related records (to ensure that associated records are upated
        on +record+'s save and visa-versa).
        """
        self.model = model
        self.record = record
        self.where_query = {}
        self.custom_where = []
        self.having_args = []
        self.join_args = []
        self._order_with = {}
        self.group_column = None
        self.limit_count = None
        self.attributes = ["id"] + list(self.model.__all_attributes__)
        self.table = Repo.table_name(self.model)

    def copy(self):
        q = Query(self.model, self.record)
        q.where_query = dict(self.where_query)
        q.custom_where = list(self.custom_where)
        q.having_args = list(self.having_args)
        q.join_args = list(self.join_args)
        q._order_with = dict(self._order_with)
        q.group_column = self.group_column
        q.attributes = list(self.attributes)
        return q

    def all(self):
        """
        Returns all records that match the query.
        """
        return self

    def find(self, id):
        """
        Find record by +id+, raising RecordNotFound if no record exists.
        """
        return self.find_by(id=id)

    def find_by(self, **kwargs):
        """
        Find first record subject to restrictions in +kwargs+, raising
        RecordNotFound if no such record exists.
        """
        result = self.where(**kwargs).first()
        if result:
            return result
        else:
            raise RecordNotFound(kwargs)

    @does_not_mutate
    def first(self, count=1):
        """
        Returns the first record in the query (sorting by id unless modified by
        `order_by`), returning None if the query has no records.
        """
        if count == 0:
            raise QueryInvalid("Count cannot be zero.")
        self.limit_count = count
        records = self._do_query().fetchall()
        if not records:
            return None
        if count == 1:
            record = records[0]
            args = dict(zip(self.attributes, record))
            return self.model.from_dict(**args)
        return self

    @does_not_mutate
    def last(self, count=1):
        """
        Returns the last record in the query (sorting by id unless modified by
        `order_by`, whereupon it reverses the order passed in `order_by`).
        Returns None if the query has no records.
        """
        if self._order_with:
            order = self._order_with.values()[0]
            order = "desc" if order == "asc" else "asc"
            order_with = self._order_with
            self._order_with = {}
            result = self.order_by(**{order_with.keys()[0]: order}).first(count)
            self._order_with = order_with
            return result
        else:
            return self.order_by(id="desc").first(count)

    @does_not_mutate
    def order_by(self, **kwargs):
        """
        Orders the query by the key passed in +kwargs+. Only pass one key, as
        it cannot sort by multiple columns at once. Raises QueryInvalid if this
        method is called when there is already a custom order (i.e. this
        method was already called on this query). Analog to "ORDER BY" in SQL.
        """
        # Only get one thing from kwargs (we can only order by one thing...)
        if self._order_with:
            raise QueryInvalid("Cannot order by more than one column")
        self._order_with = dict([kwargs.popitem()])
        return self

    @does_not_mutate
    def where(self, *custom_restrictions, **restrictions):
        """
        Restricts the records to the query subject to the passed
        +restrictions+. Analog to "WHERE" in SQL. Can pass multiple
        restrictions, and can invoke this method multiple times per query.
        """
        for attr, value in restrictions.items():
            self.where_query[attr] = value
        if custom_restrictions:
            self.custom_where.append(tuple(custom_restrictions))
        return self

    @does_not_mutate
    def joins(self, table):
        """
        Analog to "INNER JOIN" in SQL on the passed +table+. Use only once
        per query.
        """

        def do_join(table, model):
            while model is not associations.model_from_name(table):
                # ex) Category -> Forum -> Thread -> Post
                # Category: {"posts": "forums"}
                # Forum: {"posts": "threads"}
                # Thread: {"posts": None}
                # >>> Category.joins("posts")
                # => [
                #       {'table': 'forums', 'on': ['category_id', 'id']}
                #       {'table': 'threads', 'on': ['forum_id', 'id']}
                #       {'table': 'posts', 'on': ['thread_id', 'id']}
                #    ]
                if table in associations.associations_for(model):
                    # This to next: one-many (they have the fk)
                    # If associations.associations_for(model)[table] is None, then this is
                    # terminal (i.e. table is the FINAL association in the
                    # chain)
                    next_level = associations.associations_for(model)[table] or table
                    next_model = associations.model_from_name(next_level)
                    foreign_key = associations.foreign_keys_for(model).get(
                        next_level,
                        inflector.foreignKey(model.__name__))
                    yield {'table': next_level, 'on': [foreign_key, 'id']}
                else:
                    # One-One or Many-One
                    # singular table had better be in associations.associations_for(model)
                    singular = inflector.singularize(table)
                    next_level = associations.associations_for(model)[singular] or singular
                    next_model = associations.model_from_name(next_level)
                    this_table_name = Repo.table_name(model)
                    foreign_key = associations.foreign_keys_for(model).get(
                        next_level,
                        inflector.foreignKey(model.__name__))
                    if associations.model_has_foreign_key_for_table(table,
                                                                    model):
                        # we have the foreign key
                        order = ['id', foreign_key]
                    else:
                        # They have the foreign key
                        order = [foreign_key, 'id']
                    yield {'table': inflector.pluralize(next_level), 'on': order}
                model = next_model

        self.join_args = list(do_join(table, self.model))
        return self

    @does_not_mutate
    def group(self, column):
        """
        Applies a `GROUP BY` clause to the SQL generated by the Repo.
        """
        self.group_column = column
        return self

    @does_not_mutate
    def having(self, *conditions):
        """
        SQL uses the `HAVING` clause to specify conditions on the `GROUP BY`
        fields. Similarly, `having` allows specification of +conditions+ on
        fields chosen by `group`.

        >>> Order.all().group("date(created_at)").having("sum(price) > ?", 10)
        """
        self.having_args.append(tuple(conditions))
        return self

    @does_not_mutate
    def select(self, *fields):
        """
        Limit the number of fields accessed by SQL to those passed in
        +fields+.
        """
        self.attributes = fields
        return self

    def _query_repo(self):
        repo = Repo(self.table)
        if self.where_query or self.custom_where:
            repo = repo.where(self.custom_where, **self.where_query)
        if self.join_args:
            repo = repo.inner_join(*self.join_args)
        if self._order_with:
            repo = repo.order_by(**self._order_with)
        if self.group_column:
            repo = repo.group_by(self.group_column)
        if self.having_args:
            repo = repo.having(self.having_args)
        if self.limit_count:
            repo = repo.limit(self.limit_count)
        return repo

    def _do_query(self):
        return self._query_repo().select(*self.attributes)

    def __iter__(self):
        result = self._do_query().fetchall()
        for record in result:
            args = dict(zip(self.attributes, record))
            yield self.model.from_dict(**args)

    def __len__(self):
        result = self._query_repo().count()
        return result.fetchone()[0]

    def __repr__(self):
        return "<{name} {records}>".format(
            name="lazy_record.Query",
            records=list(self)
        )

    def create(self, **attributes):
        """
        Creates a new record suject to the restructions in the query and with
        the passed +attributes+. Operates using `build`.
        """
        record = self.build(**attributes)
        record.save()
        return record

    def build(self, **kwargs):
        """
        Builds a new record subject to the restrictions in the query.
        Will build intermediate (join table) records as needed, and links them
        to the returned record so that they are saved when the returned record
        is.
        """
        build_args = dict(self.where_query)
        build_args.update(kwargs)
        record = self.model(**record_args(build_args))
        if self.join_args:
            # EXAMPLE:
            # Say we have a many-to-many relation like so:
            #   Post -> Tagging -> Tag
            # If we are adding a tag to a post, we need to create the tagging
            # in addition to the tag. To do this, we add a "related record" to
            # the tag which is the tagging. By building the previous record,
            # if it exists, we can recursively build a long object tree.
            # That is, when we do post.tags(), we get a query that looks like:
            # Query(Tag).joins("taggings").where(taggings={"post_id":post.id})
            # The call to "build" then creates a tag using kwargs and the where
            # constraint, and creates a tagging that looks like:
            # Tagging(post_id = post.id) and adds it to the tag's related
            # records. The tagging's tag_id is added once the tag is saved
            # which is the first time it gets an id
            # Join args will never have just one element
            next_to_build = getattr(self.record, self.join_args[-2]['table']).build()
            record._related_records.append(next_to_build)
        return record

    def append(self, record):
        """
        Adds the passed +record+ to satisfy the query. Only intended to be
        used in conjunction with associations (i.e. do not use if self.record
        is None).

        Intended use case (DO THIS):

        post.comments.append(comment)

        NOT THIS:

        Query(Post).where(content="foo").append(post)
        """
        if self.record:
            self._validate_record(record)
            if self.join_args:
                # As always, the related record is created when the primary
                # record is saved
                build_args = dict(self.where_query)
                # The +final_join+ is what connects the record chain to the
                # passed +record+
                final_join = self.join_args[-2]
                # don't need to worry about one-to-many through because
                # there is not enough information to find or create the
                # joining record
                # i.e. in the Forum -> Thread -> Post example
                # forum.posts.append(post) doesn't make sense since there
                # is no information about what thread it will be attached to
                # Thus, this only makes sense on many-to-many. BUT we still
                # have to consider the case where there is a one-many-many
                # To make that work, we need to treat this like when doing
                # building
                joining_relation = getattr(self.record, final_join['table'])
                # Uses the lookup info in the join to figure out what ids to
                # set, and where to get the id value from
                joining_args = {final_join['on'][0]:
                                getattr(record, final_join['on'][1])}
                build_args.update(joining_args)
                joining_record = joining_relation.build(**build_args)
                self.record._related_records.append(joining_record)
            else:
                # Add our id to their foreign key so that the relationship is
                # created
                setattr(record,
                        foreign_key(record, self.record),
                        self.record.id)
                # Add to the list of related records so that it is saved when
                # we are
                self.record._related_records.append(record)

    def delete(self, record):
        """
        Removes a record from an query. Does not destroy the passed +record+,
        but marks any joining records for destruction (in the case of a
        many-to-many relationship) or sets the foreign key of +record+ to None
        (in the case of a one-to-many relationship). Only intended to be
        used in conjunction with associations (i.e. do not use if self.record
        is None).
        """
        # note: does (and should) not delete or destroy the record
        if self.record:
            self._validate_record(record)
            if self.join_args:
                # Need to find out who has the foreign key
                # If record has it, set to None, then done.
                # If one level up has it, mark the record for destruction
                final_table = self.join_args[0]['table']
                if final_table in associations.associations_for(self.model):
                    # +record+ does not have the foreign key
                    # Find the record one level up, then mark for destruction
                    related_class = associations.model_from_name(
                        final_table)
                    related_record = related_class.find_by(
                        **self._related_args(record, related_class))
                    # mark the joining record to be destroyed the primary is saved
                    self.record._delete_related_records.append(related_record)
                else:
                    # We have the foreign key
                    # Look up in the foreign key table, bearing in mind that
                    # this is a belongs_to, so the entry will be singular,
                    # whereas the table name is plural (we need to remove the
                    # 's' at the end)
                    key = associations.foreign_keys_for(self.model
                            )[inflector.singularize(final_table)]
                    # Set the foreign key to None to deassociate
                    setattr(record, key, None)
            else:
                setattr(record, foreign_key(record, self.record), None)
                # Ensure that the change is persisted on save
                self.record._related_records.append(record)

    def _related_args(self, record, related_class):
        # Both records are already persisted (have ids), so we can
        # set up the relating record fully now. One of the ids comes
        # from the constraint on the query, the other comes from
        # the foreign key logic below:
        # What we do is we get the singular table name of the record.
        # With that, we can look into the related class description for
        # the correct foreign key, which is set to the passed record's
        # id.
        record_class_name = inflector.singularize(Repo.table_name(record.__class__))
        related_args = self.where_query.get(Repo.table_name(related_class), {})
        related_key = associations.foreign_keys_for(related_class)[record_class_name]
        related_args[related_key] = record.id
        return related_args

    def _validate_record(self, record):
        if record.__class__ != self.model:
            raise AssociationTypeMismatch(
                "Expected record of type {expected}, got {actual}.".format(
                    expected=self.model.__name__,
                    actual=record.__class__.__name__
                ))

    def __getattr__(self, attr):
        # Check to see if there is a scope with this name on the queried model
        if attr in self.model.__scopes__:
            # Grab the scope lambda
            scope = self.model.get_scope(attr)
            # Bind the lambda to self so it acts like a method
            bound_scope = types.MethodType(scope, self)
            # Define the method on self
            setattr(self, attr, bound_scope)
            # Try again. This time, it will find the attribute, so __getattr__
            # will not get called again
            return getattr(self, attr)
        else:
            # Not a scope: resume standard attribute lookup
            return self.__getattribute__(attr)

    class __metaclass__(type):
        def __repr__(self):
            return "<class 'lazy_record.Query'>"


def foreign_key(local, foreign):
    local_class = local.__class__
    foreign_class = foreign.__class__
    return associations.foreign_keys_for(local_class
        )[inflector.singularize(Repo.table_name(foreign_class))]

def record_args(arg_dict):
    return {key: value
            for key, value in arg_dict.items()
            if type(value) is not dict}

# Here to prevent circular import loop
from lazy_record.errors import *
import lazy_record.associations as associations



import unittest
import mock
import sys
import os
sys.path.insert(0, os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
sys.path.append(os.path.join(
    os.path.dirname(os.path.abspath(os.path.dirname(__file__))),
    "lazy_record"))
import base


class Cantaloupe(base.Base):
    __scopes__ = {
        "ripe": lambda query: query.where(ripe=1)
    }

class Watermelon(base.Base):
    pass

class TestScopes(unittest.TestCase):

    @mock.patch("base.Query")
    def test_defines_class_method_on_model(self, Query):
        Cantaloupe.ripe()
        Query.assert_called_with(Cantaloupe)
        query = Query.return_value
        query.where.assert_called_with(ripe=1)

    def test_class_method_has_reasonable_signature(self):
        self.assertEqual(Cantaloupe.ripe.__name__, "<scope>ripe")

    @mock.patch("base.Query.where")
    def test_defines_method_on_classes_queries(self, where):
        query = base.Query(Cantaloupe)
        query.ripe()
        where.assert_called_with(ripe=1)

    @mock.patch("base.Query.where")
    def test_does_not_define_method_on_other_queries(self, where):
        query = base.Query(Cantaloupe)
        query.ripe()
        query2 = base.Query(Watermelon)
        with self.assertRaises(AttributeError) as e:
            query2.ripe()
        self.assertEqual(e.exception.message,
                         "'Query' object has no attribute 'ripe'")

    def test_does_not_impede_attribute_lookup_on_class(self):
        with self.assertRaises(AttributeError) as e:
            Cantaloupe.turnip
        self.assertEqual(e.exception.message,
                         "'Cantaloupe' has no attribute 'turnip'")

if __name__ == '__main__':
    unittest.main()



"""
Module of common validation functions used to validate records.
"""

__all__ = ["present", "unique", "absent", "length", "validation"]

class validation(object):

    def __init__(self, fun):
        self.fun = fun

    def __call__(self, record, name=None):
        return self.fun(record, name or self.name)

@validation
def present(record, name):
    return bool(getattr(record, name))

@validation
def absent(record, name):
    return not present(record, name)

@validation
def unique(record, name):
    model = record.__class__
    others = model.where("id IS NOT ?", record.id
                 ).where("{} == ?".format(name), getattr(record, name))
    return len(list(others)) == 0

def length(within):
    @validation
    def length_validator(record, name):
        return len(getattr(record, name)) in within
    return length_validator



import unittest
import mock
import sys
import os
sys.path.insert(0, os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.abspath(os.path.dirname(__file__))),
    "lazy_record"))
import validations


class TestValidations(unittest.TestCase):

    def test_validates_presence_when_present(self):
        validator = validations.present
        validator.name = "age"
        self.assertTrue(validator(mock.Mock(age="me")))
        self.assertTrue(validator(mock.Mock(age=17)))
        self.assertTrue(validator(mock.Mock(age=True)))

    def test_validates_presence_when_not_present(self):
        validator = validations.present
        validator.name = "age"
        self.assertFalse(validator(mock.Mock(age="")))
        self.assertFalse(validator(mock.Mock(age=False)))
        self.assertFalse(validator(mock.Mock(age=None)))

    def test_validates_absence_when_present(self):
        validator = validations.absent
        validator.name = "age"
        self.assertFalse(validator(mock.Mock(age=17)))
        self.assertFalse(validator(mock.Mock(age=True)))

    def test_validates_absence_when_not_present(self):
        validator = validations.absent
        validator.name = "age"
        self.assertTrue(validator(mock.Mock(age="")))
        self.assertTrue(validator(mock.Mock(age=False)))
        self.assertTrue(validator(mock.Mock(age=None)))

    def test_validates_uniqueness_when_unique(self):
        validator = validations.unique
        validator.name = "age"
        record = mock.Mock(age=15, __class__=mock.Mock())
        klass = record.__class__
        klass.where.return_value.where.return_value = []
        self.assertTrue(validator(record))

    def test_validates_uniqueness_when_not_unique(self):
        validator = validations.unique
        validator.name = "age"
        record = mock.Mock(age=15, __class__=mock.Mock())
        klass = record.__class__
        klass.where.return_value.where.return_value = [mock.Mock()]
        self.assertFalse(validator(record))

    def test_validates_length_when_too_short(self):
        validator = validations.length(within=range(2, 5))
        validator.name = "age"
        self.assertFalse(validator(mock.Mock(age="")))

    def test_validates_length_when_too_long(self):
        validator = validations.length(within=range(2, 5))
        validator.name = "age"
        self.assertFalse(validator(mock.Mock(age="asfafds")))

    def test_validates_length_when_correct(self):
        validator = validations.length(within=range(2, 5))
        validator.name = "age"
        self.assertTrue(validator(mock.Mock(age="abcd")))


if __name__ == '__main__':
    unittest.main()



import unittest
import sys
import os
# This way, we pick the lazy_record local even if one is installed
sys.path.insert(0, os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.abspath(os.path.dirname(__file__))),
    "lazy_record"))
from lazy_record.associations import *
import lazy_record

# INTEGRATED TEST

@has_one("other_thing", through="thing")
@has_one("thing")
@has_many("lending_tables", through="lendings")
@has_many("people", through="lendings")
@has_many("lendings")
class Book(lazy_record.Base):
    pass


@belongs_to("book")
@belongs_to("person")
@has_many("lending_tables")
class Lending(lazy_record.Base):
    pass


@has_many("things", through="books")
@has_many("books", through="lendings")
class Person(lazy_record.Base):
    pass

@belongs_to("lending")
class LendingTable(lazy_record.Base):
    pass

@has_one("other_thing")
@belongs_to("book")
class Thing(lazy_record.Base):
    pass

@belongs_to("thing")
class OtherThing(lazy_record.Base):
    pass

@has_one("end_two", through="joiner")
@has_one("joiner")
class EndOne(lazy_record.Base):
    pass

@belongs_to("end_one")
@belongs_to("end_two")
class Joiner(lazy_record.Base):
    pass

@has_one("end_one", through="joiner")
@has_one("joiner")
class EndTwo(lazy_record.Base):
    pass

test_schema = """
drop table if exists people;
create table people (
  id integer primary key autoincrement,
  created_at timestamp not null,
  updated_at timestamp not null
);
drop table if exists lendings;
create table lendings (
  id integer primary key autoincrement,
  person_id integer not null,
  book_id integer not null,
  created_at timestamp not null,
  updated_at timestamp not null
);
drop table if exists books;
create table books (
  id integer primary key autoincrement,
  created_at timestamp not null,
  updated_at timestamp not null
);
drop table if exists lending_tables;
create table lending_tables (
  id integer primary key autoincrement,
  lending_id integer,
  created_at timestamp not null,
  updated_at timestamp not null
);
drop table if exists things;
create table things (
  id integer primary key autoincrement,
  book_id integer,
  created_at timestamp not null,
  updated_at timestamp not null
);
drop table if exists other_things;
create table other_things (
  id integer primary key autoincrement,
  thing_id integer,
  created_at timestamp not null,
  updated_at timestamp not null
);
drop table if exists end_ones;
create table end_ones (
  id integer primary key autoincrement,
  created_at timestamp not null,
  updated_at timestamp not null
);
drop table if exists end_twos;
create table end_twos (
  id integer primary key autoincrement,
  created_at timestamp not null,
  updated_at timestamp not null
);
drop table if exists joiners;
create table joiners (
  id integer primary key autoincrement,
  end_one_id integer,
  end_two_id integer,
  created_at timestamp not null,
  updated_at timestamp not null
)
"""


class TestGettingRecordsThroughJoin(unittest.TestCase):

    def setUp(self):
        lazy_record.connect_db()
        lazy_record.load_schema(test_schema)
        self.person = Person.create()
        self.book = Book.create()
        self.lending = Lending.create(person_id=self.person.id,
                                      book_id=self.book.id)

    def tearDown(self):
        lazy_record.close_db()

    def test_finds_specific_records_through_many_to_many(self):
        self.assertIn(self.book.id, [b.id for b in self.person.books])

    def test_finds_records_through_many_to_many(self):
        person = Person.create()
        query = Person.joins("books")
        # Join finds the related record
        assert (self.person.id in [p.id for p in query])
        # Join does NOT find the unrelated record
        assert (person.id not in [p.id for p in query])

    def test_finds_records_through_one_to_many(self):
        person = Person.create()
        query = Person.joins("lendings")
        # Join finds the related record
        assert (self.person.id in [p.id for p in query])
        # Join does NOT find the unrelated record
        assert (person.id not in [p.id for p in query])

    def test_finds_records_through_deep_one_to_many(self):
        # import pdb; pdb.set_trace()
        lending = Lending.first()
        lending_table = LendingTable(lending_id=lending.id)
        lending_table.save()
        # import pdb; pdb.set_trace()
        self.assertIn(lending_table.id,
                      [l.id for l in self.book.lending_tables])

    def finds_records_one_deep(self):
        self.assertIn(self.lending, self.person.lendings)

class TestBuildingRecordsThroughJoin(unittest.TestCase):

    def setUp(self):
        lazy_record.connect_db()
        lazy_record.load_schema(test_schema)
        self.person = Person()
        self.person.save()

    def tearDown(self):
        lazy_record.close_db()

    def test_creates_records_and_intermediates(self):
        book = self.person.books.build()
        book.save()
        assert (book.id in [b.id for b in self.person.books])
        assert (self.person.id in [p.id for p in book.people])


class TestDestroyingRecordsThroughJoin(unittest.TestCase):

    def setUp(self):
        lazy_record.connect_db()
        lazy_record.load_schema(test_schema)
        self.person = Person()
        self.person.save()
        self.book = Book()
        self.book.save()
        Lending(person_id=self.person.id, book_id=self.book.id).save()

    def tearDown(self):
        lazy_record.close_db()

    def test_deletes_only_join_record(self):
        self.person.destroy()
        # Test that the relationship is gone
        assert len(list(self.person.books)) == 0, \
            ("Expected self.person.books to be empty, "
             "but it had count {}".format(len(list(self.person.books))))
        # test that the lending is gone
        self.assertEqual(len(list(Lending.all())), 0)
        # test that the book still exists
        assert Book.find(self.book.id).id == self.book.id

    def test_delete_on_query_unlinks_records_and_destroys_join(self):
        self.person.books.delete(self.book)
        self.person.save()
        assert len(list(self.person.books)) == 0, \
            ("Expected self.person.books to be empty, "
             "but it had count {}".format(len(list(self.person.books))))
        assert Book.find(self.book.id).id == self.book.id
        assert Person.find(self.person.id).id == self.person.id

    def test_delete_on_query_unlinks_record_on_deep_one_to_many(self):
        lending = Lending.first()
        lending_table = LendingTable(lending_id=lending.id)
        lending_table.save()
        self.book.lending_tables.delete(lending_table)
        self.book.save()
        self.assertEqual(lending_table.lending_id, None)
        # Should not destroy the lending (compare with many-to-many)
        self.assertEqual(Lending.first().id, lending.id)

    def test_raises_AssociationTypeMismatch_not_correct_type(self):
        with self.assertRaises(lazy_record.AssociationTypeMismatch) as e:
            self.person.books.delete(self.person)
        self.assertEqual(e.exception.message,
                         "Expected record of type Book, got Person.")

class TestManyThroughOne(unittest.TestCase):

    def setUp(self):
        lazy_record.connect_db()
        lazy_record.load_schema(test_schema)
        self.person = Person.create()
        self.book = Book.create()
        self.lending = Lending.create(person_id=self.person.id,
                                      book_id=self.book.id)

    def test_finds_many_through_one(self):
        thing = Thing.create(book_id=self.book.id)
        self.assertIn(thing, self.person.things)

class TestOneToOne(unittest.TestCase):

    def setUp(self):
        lazy_record.connect_db()
        lazy_record.load_schema(test_schema)
        self.book = Book.create()
        self.thing = Thing.create(book_id=self.book.id)
        self.other_thing = OtherThing.create(thing_id=self.thing.id)

    def test_sets_to_none(self):
        self.thing.other_thing = None
        self.thing.save()
        self.assertEqual(self.thing.other_thing, None)
        other_thing = OtherThing.find(self.other_thing.id)
        self.assertEqual(other_thing.thing_id, None)

class TestOneThroughOne(unittest.TestCase):

    def setUp(self):
        lazy_record.connect_db()
        lazy_record.load_schema(test_schema)
        self.book = Book.create()
        self.thing = Thing.create(book_id=self.book.id)
        self.other_thing = OtherThing.create(thing_id=self.thing.id)

    def test_gets_record(self):
        self.assertEqual(self.book.other_thing, self.other_thing)

    def test_sets_record(self):
        another_other_thing = OtherThing()
        self.book.other_thing = another_other_thing
        self.assertEqual(another_other_thing.thing_id, self.thing.id)
        self.book.save()
        self.assertEqual(another_other_thing, self.book.other_thing)

    def test_disassociates_previous_record(self):
        another_other_thing = OtherThing()
        self.book.other_thing = another_other_thing
        self.book.save()
        # Need to reload the object to see the changes
        other_thing = OtherThing.find(self.other_thing.id)
        self.assertEqual(other_thing.thing_id, None)

    def test_sets_child_to_none(self):
        self.book.other_thing = None
        self.book.save()
        # Need to reload the object to see the changes
        other_thing = OtherThing.find(self.other_thing.id)
        self.assertEqual(other_thing.thing_id, None)
        self.assertEqual(self.book.other_thing, None)

    def test_gets_child_no_belongs_to(self):
        end_1 = EndOne.create()
        end_2 = EndTwo.create()
        joiner = Joiner.create(end_one_id=end_1.id, end_two_id=end_2.id)
        self.assertEqual(end_1.end_two, end_2)
        self.assertEqual(end_2.end_one, end_1)

    def test_sets_child_no_belongs_to(self):
        end_1 = EndOne.create()
        end_2 = EndTwo.create()
        end_2_new = EndTwo.create()
        joiner = Joiner.create(end_one_id=end_1.id, end_two_id=end_2.id)
        end_1.end_two = end_2_new
        end_1.save()
        self.assertEqual(end_1.end_two, end_2_new)
        self.assertEqual(end_2.end_one, None)

    def test_sets_child_to_none_no_belongs_to(self):
        end_1 = EndOne.create()
        end_2 = EndTwo.create()
        joiner = Joiner.create(end_one_id=end_1.id, end_two_id=end_2.id)
        end_1.end_two = None
        end_1.save()
        self.assertEqual(end_1.end_two, None)
        self.assertEqual(end_2.end_one, None)

class TestAddsRecords(unittest.TestCase):

    def setUp(self):
        lazy_record.connect_db()
        lazy_record.load_schema(test_schema)
        self.person = Person()
        self.person.save()
        self.book = Book()
        self.book.save()

    def tearDown(self):
        lazy_record.close_db()

    def test_appends_records_without_join(self):
        lending = Lending(person_id=self.person.id)
        self.book.lendings.append(lending)
        self.book.save()
        assert (self.book.id in [b.id for b in self.person.books])
        assert (self.person.id in [p.id for p in self.book.people])

    def test_adds_intermediaries(self):
        self.person.books.append(self.book)
        self.person.save()
        assert (self.book.id in [b.id for b in self.person.books])
        assert (self.person.id in [p.id for p in self.book.people])

    def test_raises_AssociationTypeMismatch_not_correct_type(self):
        with self.assertRaises(lazy_record.AssociationTypeMismatch) as e:
            self.person.books.append(self.person)
        self.assertEqual(e.exception.message,
                         "Expected record of type Book, got Person.")

    def test_sets_record_in_init(self):
        lending = Lending(person=self.person, book=self.book)
        lending.save()
        self.assertEqual(lending.person_id, self.person.id)
        self.assertEqual(lending.book_id, self.book.id)

if __name__ == '__main__':
    unittest.main()



from lazy_record.errors import *
import lazy_record.validations as validators

class Validations(object):
    __validates__ = {}

    def is_valid(self, attrs = None):
        """
        Validate an object against the __validates__ class variable.
        Returns True if valid, and False if any validations fail. If passed
        (as a dict), +attrs+ will be populated with the invalid attributes.
        """
        if attrs is None:
            reason = {}
        else:
            reason = attrs
        valid = True
        for attr, validation in self.__class__.__validates__.items():
            if validation.__class__ == validators.validation:
                validation.name = attr
            if not validation(self):
                reason[attr] = getattr(self, attr)
                valid = False
        return valid

    def validate(self):
        """
        Validate an object against the __validates__ class variable.
        Returns None on success, and raises RecordInvalid if validations fail.
        """
        reason = {}
        if not self.is_valid(attrs=reason):
            raise RecordInvalid(reason)



from query import Query
from base import Base
from errors import *
from typecasts import *


__author__ = "Chase Conklin"


def connect_db(database_name=":memory:"):
    """
    Connect lazy_record to the database at the path specified in
    +database_name+.
    """
    db = repo.Repo.connect_db(database_name)
    base.Repo.db = db
    query.Repo.db = db


def close_db():
    """
    Close the connection to the database opened in `connect_db`
    """
    db = repo.Repo.db
    if db is not None:
        db.close()
    repo.Repo.db = None
    base.Repo.db = None
    query.Repo.db = None

def load_schema(schema):
    """
    Load a schema file with path +schema+ into the database. Assumes that
    there exists an active database connection.
    """
    with repo.Repo.db:
        repo.Repo.db.executescript(schema)



class RecordNotFound(Exception):
    pass


class RecordInvalid(Exception):
    pass


class QueryInvalid(Exception):
    pass


class AssociationTypeMismatch(Exception):
    pass

class MissingAttributeError(Exception):
    pass

class AssociationForbidden(Exception):
    pass



import unittest
import mock
import sys
import os
sys.path.append(os.path.join(
    os.path.dirname(os.path.abspath(os.path.dirname(__file__))),
    "lazy_record"))
from query import Query, Repo
import datetime
import query


class TunaCasserole(object):
    __all_attributes__ = {"my_attr": int, "created_at": int, "updated_at":int}
    __scopes__ = {}

    def __init__(self, **kwargs):
        self._related_records = []
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def from_dict(TunaCasserole, **kwargs):
        return kwargs

class MyRelations(object):
    pass

@mock.patch("query.Repo")
class TestQuery(unittest.TestCase):

    def test_instantiates_repo(self, Repo):
        Repo.table_name.return_value = "tuna_casseroles"
        list(Query(TunaCasserole).all())
        Repo.table_name.assert_called_with(TunaCasserole)
        Repo.assert_called_with("tuna_casseroles")

    def test_makes_query_for_all_records(self, Repo):
        Repo.table_name.return_value = "tuna_casseroles"
        list(Query(TunaCasserole).all())
        repo = Repo.return_value
        repo.select.assert_called_with("id", "created_at",
                                       "updated_at", "my_attr")

    def test_constructs_object_with_information(self, Repo):
        repo = Repo.return_value
        fetchall_return = [(2, 33, 15)]
        fetchall = mock.Mock(return_value=fetchall_return)
        select_mock = mock.Mock(fetchall=fetchall)
        repo.select.return_value = mock.Mock(fetchall=fetchall)
        self.assertEqual(list(Query(TunaCasserole).all())[0],
            {"id": 2, "updated_at": 15, "created_at": 33})

    def test_where_restricts_query(self, Repo):
        list(Query(TunaCasserole).where(my_attr=5))
        repo = Repo.return_value
        repo.where.assert_called_with([], my_attr=5)
        repo.where.return_value.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")

    def test_where_allows_all_after(self, Repo):
        list(Query(TunaCasserole).where(my_attr=5).all())
        repo = Repo.return_value
        repo.where.assert_called_with([], my_attr=5)
        repo.where.return_value.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")

    def test_where_allows_chaining(self, Repo):
        list(Query(TunaCasserole).where(my_attr=5).where(id=7))
        repo = Repo.return_value
        repo.where.assert_called_with([], my_attr=5, id=7)
        repo.where.return_value.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")

    def test_where_allows_arbitrary_queries(self, Repo):
        list(Query(TunaCasserole).where("my_attr == ?", 5))
        repo = Repo.return_value
        repo.where.assert_called_with([("my_attr == ?", 5)])
        repo.where.return_value.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")

    def test_where_with_arbitrary_queries_allows_chaining(self, Repo):
        list(Query(TunaCasserole).where("my_attr > ?", 5).where(
            "my_attr < ?", 10))
        repo = Repo.return_value
        repo.where.assert_called_with([("my_attr > ?", 5),
                                       ("my_attr < ?", 10)])
        repo.where.return_value.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")

    def test_allows_ordering(self, Repo):
        list(Query(TunaCasserole).order_by(id="desc").all())
        repo = Repo.return_value
        repo.order_by.assert_called_with(id="desc")
        order = repo.order_by.return_value
        order.select.assert_called_with("id", "created_at",
                                        "updated_at", "my_attr")

    def test_raises_on_multiple_orders(self, Repo):
        with self.assertRaises(query.QueryInvalid):
            q = Query(TunaCasserole).order_by(id="desc").order_by(id="asc")
            list(q.all())

    def test_gets_first_record(self, Repo):
        record = Query(TunaCasserole).where(my_attr=5).where(id=7).first()
        self.assertEqual({}, record)
        repo = Repo.return_value
        repo.where.assert_called_with([], my_attr=5, id=7)
        where = repo.where.return_value
        where.limit.assert_called_with(1)
        limit = where.limit.return_value
        limit.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")
        select = limit.select.return_value
        select.fetchall.assert_called_once_with()

    def test_first_returns_None_if_no_first_record(self, Repo):
        repo = Repo.return_value
        where = repo.where.return_value
        where.limit.return_value.select.return_value.fetchall.return_value = []
        record = Query(TunaCasserole).where(my_attr=5).where(id=7).first()
        self.assertEqual(None, record)

    def test_first_gets_first_few_records(self, Repo):
        records = Query(TunaCasserole).where(my_attr=5).where(id=7).first(5)
        repo = Repo.return_value
        repo.where.assert_called_with([], my_attr=5, id=7)
        where = repo.where.return_value
        where.limit.assert_called_with(5)
        limit = where.limit.return_value
        limit.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")
        select = limit.select.return_value
        select.fetchall.assert_called_once_with()

    def test_first_raises_if_count_is_zero(self, Repo):
        with self.assertRaises(query.QueryInvalid):
            Query(TunaCasserole).where(my_attr=5).where(id=7).first(0)

    def test_last_raises_if_count_is_zero(self, Repo):
        with self.assertRaises(query.QueryInvalid):
            Query(TunaCasserole).where(my_attr=5).where(id=7).last(0)

    def test_gets_last_record(self, Repo):
        record = Query(TunaCasserole).where(my_attr=5).where(id=7).last()
        self.assertEqual({}, record)
        repo = Repo.return_value
        repo.where.assert_called_with([], my_attr=5, id=7)
        where = repo.where.return_value
        where.order_by.assert_called_with(id="desc")
        order = where.order_by.return_value
        order.limit.assert_called_with(1)
        limit = order.limit.return_value
        limit.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")
        select = limit.select.return_value
        select.fetchall.assert_called_once_with()

    def test_last_gets_last_few_records(self, Repo):
        record = Query(TunaCasserole).where(my_attr=5).where(id=7).last(5)
        repo = Repo.return_value
        repo.where.assert_called_with([], my_attr=5, id=7)
        where = repo.where.return_value
        where.order_by.assert_called_with(id="desc")
        order = where.order_by.return_value
        order.limit.assert_called_with(5)
        limit = order.limit.return_value
        limit.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")
        select = limit.select.return_value
        select.fetchall.assert_called_once_with()

    def test_last_returns_None_if_no_last_record(self, Repo):
        repo = Repo.return_value
        where = repo.where.return_value
        order = where.order_by.return_value
        limit = order.limit.return_value
        select = limit.select.return_value
        select.fetchall.return_value = []
        record = Query(TunaCasserole).where(my_attr=5).where(id=7).last()
        self.assertEqual(None, record)

    def test_gets_last_record_with_existing_sort(self, Repo):
        r = Query(TunaCasserole).where(my_attr=5).order_by(id="desc").last()
        self.assertEqual({}, r)
        repo = Repo.return_value
        repo.where.assert_called_with([], my_attr=5)
        where = repo.where.return_value
        where.order_by.assert_called_with(id="asc")
        order = where.order_by.return_value
        order.limit.assert_called_with(1)
        limit = order.limit.return_value
        limit.select.assert_called_with(
            "id", "created_at", "updated_at", "my_attr")
        select = limit.select.return_value
        select.fetchall.assert_called_once_with()

    def test_gets_last_record_with_existing_sort_with_no_record(self, Repo):
        repo = Repo.return_value
        where = repo.where.return_value
        order = where.order_by.return_value
        limit = order.limit.return_value
        select = limit.select.return_value
        select.fetchall.return_value = []
        r = Query(TunaCasserole).where(my_attr=5).order_by(id="desc").last()
        self.assertEqual(None, r)

    def test_builds_simple_related_records(self, Repo):
        record = Query(TunaCasserole).where(my_attr=11).build()
        self.assertEqual(record.my_attr, 11)

    def test_builds_simple_related_records_with_args(self, Repo):
        record = Query(TunaCasserole).where(my_attr=11).build(my_attr=12)
        self.assertEqual(record.my_attr, 12)

    @mock.patch("query.associations.foreign_keys_for")
    def test_unrelates_records(self, fkf, Repo):
        fkf.return_value = {'tuna_casserole': 'tuna_casserole_id'}
        Repo.table_name.return_value = "tuna_casseroles"
        t = TunaCasserole()
        t2 = TunaCasserole(tuna_casserole_id=15)
        Query(TunaCasserole, record=t).delete(t2)
        self.assertEqual(t2.tuna_casserole_id, None)

    def test_displays_as_empty_query(self, Repo):
        Repo.return_value.select.return_value.fetchall.return_value = []
        self.assertEqual(repr(Query(TunaCasserole)), "<lazy_record.Query []>")

    def test_displays_as_query_with_records(self, Repo):
        Repo.return_value.select.return_value.fetchall.return_value = [
            (1, 7, datetime.datetime(2016, 1, 1))]
        # Recall that TunaCasserole overrides #from_dict to return
        # 'mytestvalue' so that is what it will repr as
        self.assertEqual(repr(Query(TunaCasserole)),
                     "<lazy_record.Query [{'created_at': 7, 'id': 1, "
                     "'updated_at': datetime.datetime(2016, 1, 1, 0, 0)}]>")

    def test_class_displays_as_though_it_was_in_lazy_record(self, Repo):
        self.assertEqual(repr(Query), "<class 'lazy_record.Query'>")

    def tests_inclusion(self, Repo):
        repo = Repo.return_value
        fetchall_return = [(15, 2, 33)]
        fetchall = mock.Mock(return_value=fetchall_return)
        select_mock = mock.Mock(fetchall=fetchall)
        repo.select.return_value = mock.Mock(fetchall=fetchall)
        self.assertIn({'created_at': 2, 'id': 15, 'updated_at': 33},
                      Query(TunaCasserole).all())

    def test_len_invokes_SQL_count_function(self, Repo):
        repo = Repo.return_value
        len(Query(TunaCasserole).all())
        repo.count.assert_called_with

    def test_returns_SQL_count(self, Repo):
        fetchone = Repo.return_value.count.return_value.fetchone
        fetchone.return_value = (2276,)
        self.assertEqual(len(Query(TunaCasserole).all()), 2276)

    def test_where_with_list_uses_sql_in(self, Repo):
        repo = Repo.return_value
        list(Query(TunaCasserole).where(name=["foo", "bar", "baz"]))
        repo.where.assert_called_with([], name=["foo", "bar", "baz"])

    def test_group_delegates_to_repo(self, Repo):
        repo = Repo.return_value
        list(Query(TunaCasserole).group("name"))
        repo.group_by.assert_called_with("name")

    def test_having_delegates_to_repo(self, Repo):
        repo = Repo.return_value
        list(Query(TunaCasserole).group("name").having("amount > ?", 15))
        repo.group_by.return_value.having.assert_called_with([("amount > ?", 15)])

    def test_select_limits_selected_attributes(self, Repo):
        repo = Repo.return_value
        list(Query(TunaCasserole).select("name"))
        repo.select.assert_called_with("name")

    def test_creates_invokes_build(self, Repo):
        query = Query(TunaCasserole).where(my_attr=11)
        query.build = mock.Mock(name="build")
        query.create(name="foo")
        query.build.assert_called_with(name="foo")

    def test_create_saves_record(self, Repo):
        record = mock.Mock(name="record")
        query = Query(TunaCasserole).where(my_attr=11)
        query.build = mock.Mock(name="build", return_value=record)
        query.create(name="foo")
        record.save.assert_called_with()

    def test_returns_saved_record(self, Repo):
        record = mock.Mock(name="record")
        query = Query(TunaCasserole).where(my_attr=11)
        query.build = mock.Mock(name="build", return_value=record)
        self.assertEqual(query.create(name="foo"), record)

    def test_find_records_when_exists(self, Repo):
        repo = Repo.return_value
        fetchone_return = {"id": 5, "my_attr": 15, "created_at": 33}
        fetchall = mock.Mock(return_value=[fetchone_return])
        repo.where.return_value.limit.return_value.select.return_value = mock.Mock(
            fetchall=fetchall)
        Query(TunaCasserole).find(5)
        repo.where.assert_called_with([], id=5)

    def test_find_raises_when_no_record(self, Repo):
        repo = Repo.return_value
        fetchall = mock.Mock(return_value=[])
        repo.where.return_value.select.return_value = mock.Mock(
            fetchall=fetchall)
        with self.assertRaises(query.RecordNotFound):
            Query(TunaCasserole).find(5)
        repo.where.assert_called_with([], id=5)

    def test_allows_finding_of_records_by_attribute(self, Repo):
        repo = Repo.return_value
        fetchone_return = {"id": 5, "my_attr": 15, "created_at": 33}
        fetchall = mock.Mock(return_value=[fetchone_return])
        repo.where.return_value.limit.return_value.select.return_value = mock.Mock(
            fetchall=fetchall)
        Query(TunaCasserole).find_by(name="foo")
        repo.where.assert_called_with([], name="foo")

    def test_raises_when_find_by_finds_nothing(self, Repo):
        repo = Repo.return_value
        fetchall = mock.Mock(return_value=[])
        repo.where.return_value.limit.return_value.select.return_value = mock.Mock(
            fetchall=fetchall)
        with self.assertRaises(query.RecordNotFound):
            Query(TunaCasserole).find_by(name="foo")
        repo.where.assert_called_with([], name="foo")


class TestRepeatedQueries(unittest.TestCase):

    def test_where_does_not_mutate_query(self):
        with mock.patch("query.Repo") as Repo:
            q = Query(TunaCasserole).where(my_attr=5)
            q.where(my_attr=3)
            list(q)
            repo = Repo.return_value
            repo.where.assert_called_with([], my_attr=5)
            repo.where.return_value.select.assert_called_with(
                "id", "created_at", "updated_at", "my_attr")

    def test_gets_last_record(self):
        q = Query(TunaCasserole).where(my_attr=5).where(id=7)
        with mock.patch("query.Repo") as Repo:
            repo = Repo.return_value
            record = q.last()
        with mock.patch("query.Repo") as Repo:
            repo = Repo.return_value
            record_2 = q.first()
            self.assertEqual({}, record_2)
            repo.where.assert_called_with([], my_attr=5, id=7)
            where = repo.where.return_value
            where.order_by.assert_not_called()
            where.limit.assert_called_with(1)
            limit = where.limit.return_value
            limit.select.assert_called_with(
                "id", "created_at", "updated_at", "my_attr")
            limit.select.return_value.fetchall.assert_called_once_with()

    def test_ordering_does_not_mutate(self):
        with mock.patch("query.Repo") as Repo:
            q = Query(TunaCasserole)
            q.order_by(id="asc")
            list(q.order_by(id="desc"))
            repo = Repo.return_value
            repo.order_by.assert_called_with(id="desc")
            order = repo.order_by.return_value
            order.select.assert_called_with("id", "created_at",
                                            "updated_at", "my_attr")


if __name__ == '__main__':
    unittest.main()



import sys
import os
sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
from query import Query
from repo import Repo
import datetime
from lazy_record.errors import *
import lazy_record.typecasts as typecasts
from validations import Validations
import lazy_record.associations as associations
from itertools import chain
from inflector import Inflector, English

inflector = Inflector(English)


class Base(Validations):
    __attributes__ = {}
    __dependents__ = []
    __scopes__ = {}

    def __init__(self, **kwargs):
        """
        Instantiate a new object, mass-assigning the values in +kwargs+.
        Cannot mass-assign id or created_at.
        """
        if set(["id", "created_at", "updated_at"]) & set(kwargs):
            raise AttributeError("Cannot set 'id', 'created_at', "
                                 "or 'updated_at'")
        for attr in self.__class__.__all_attributes__:
            setattr(self, "_" + attr, None)
        self.update(**kwargs)
        self._id = None
        self.__table = Repo.table_name(self.__class__)
        self._related_records = []
        self._delete_related_records = []

    def __getattr__(self, attr):
        """
        Get and cast +attr+ according to the __attributes__ class variable.

        ex)
        >>> class MyRecord(lazy_record.Base)
        ...     __attributes__ = {
        ...         "my_val": int
        ...     }
        ...
        >>> record = MyRecord(my_val='11')
        >>> record.my_val
        11
        """
        def identity(val): return val
        attr_dict = self.__class__.__all_attributes__
        if attr in attr_dict or attr == "id":
            try:
                value = self.__getattribute__("_" + attr)
                if value is not None:
                    return attr_dict.get(attr, identity)(value)
                else:
                    return None
            except AttributeError:
                raise MissingAttributeError(
                    "'{}' object has no attribute '{}'".format(
                    self.__class__.__name__, attr))
        else:
            return self.__getattribute__(attr)

    def __setattr__(self, name, value):
        """
        Cast +value+ according to the __attributes__ class variable, then set
        the attribute +name+ to the casted value.

        ex)
        >>> class MyRecord(lazy_record.Base)
        ...     __attributes__ = {
        ...         "my_val": int
        ...     }
        ...
        >>> record = MyRecord()
        >>> record.my_val = '11'
        >>> record._my_val # Don't actually do this in production code.
        11
        """
        if name in ("id", "created_at", "updated_at"):
            raise AttributeError("Cannot set '{}'".format(name))
        elif name in self.__class__.__attributes__:
            if value is not None:
                setattr(self, "_" + name,
                        self.__class__.__attributes__[name](value))
            else:
                setattr(self, "_" + name, None)
        else:
            super(Base, self).__setattr__(name, value)

    @classmethod
    def from_dict(cls, **kwargs):
        """
        Construct an object an mass-assign its attributes using +kwargs+,
        ignoring all protections of id and created_at. Intended for
        constructing objects already present in the database (i.e for use by
        methods such as find or within Query).
        """
        # Create the object
        obj = cls()
        # By default, objects are initialized with attribute values of None
        # We need to clear those out so that we get AttributeError on access
        for attr in cls.__all_attributes__:
            delattr(obj, "_" + attr)
        del obj._id
        # Set the attributes that were passed
        for attr, val in kwargs.items():
            setattr(obj, "_" + attr, val)
        return obj

    def update(self, **kwargs):
        """
        Mass-assign the attributes in +kwargs+ to the object, preventing
        attributes not in __attributes__ from being set.
        """
        for attr, val in kwargs.items():
            if attr in (list(self.__class__.__attributes__) + \
                        list(associations.associations_for(self.__class__))):
                setattr(self, attr, val)

    def delete(self):
        """
        Delete this record without deleting any dependent or child records.
        This can orphan records, so use with care.
        """
        if self.id:
            with Repo.db:
                Repo(self.__table).where(id=self.id).delete()

    def _do_destroy(self):
        Repo(self.__table).where(id=self.id).delete()
        for dependent in set(self.__class__.__dependents__):
            if dependent == inflector.singularize(dependent):
                child = getattr(self, dependent)
                if child:
                    child._do_destroy()
            else:
                for record in (getattr(self, dependent) or []):
                    record._do_destroy()

    def destroy(self):
        """
        Delete this record, while also destroying (i.e. calling destroy) on
        all dependents and children.
        """
        if self.id:
            with Repo.db:
                self._do_destroy()

    def _do_save(self):
        self.validate()
        self._updated_at = datetime.datetime.today()
        if self.id:
            attrs = list(self.__class__.__all_attributes__)
            data = {attr: getattr(self, "_" + attr) for attr in attrs}
            Repo(self.__table).where(id=self.id).update(**data)
        else:
            attrs = list(self.__class__.__all_attributes__)
            self._created_at = datetime.datetime.today()
            data = {attr: getattr(self, "_" + attr) for attr in attrs}
            self.__id = int(Repo(self.__table).insert(**data))

    def _finish_save(self):
        if not self.id:
            self._id = self.__id
        for record in self._related_records:
            if not record.id:
                record._id = record.__id
        self._related_records = []

    def save(self):
        """
        Save a record to the database, creating it if needed, updating it
        otherwise. Also saves related records (children and dependents) as
        needed.
        """
        with Repo.db:
            self._do_save()
            our_name = inflector.singularize(Repo.table_name(self.__class__))
            for record in self._related_records:
                if not self._id:
                    related_key = associations.foreign_keys_for(
                        record.__class__)[our_name]
                    setattr(record, related_key, self.__id)
                record._do_save()
            for record in self._delete_related_records:
                record._do_destroy()
        self._finish_save()

    def __cmp__(self, other):
        """
        Compare to other records.
        """
        if self is other:
            return 0
        elif self.__class__ != other.__class__:
            return cmp(self.__class__, other.__class__)
        elif self.id == None:
            return 1
        elif other.id == None:
            return -1
        else:
            return cmp(self.id, other.id)

    def __int__(self):
        """
        Cast record to int (returns id).
        """
        if self.id:
            return int(self.id)
        else:
            return 0

    def __repr__(self):

        def gettimestamp(obj, attr):
            if getattr(obj, attr):
                return str(getattr(obj, attr).replace(microsecond=0))

        return "{}({})".format(
            self.__class__.__name__,
            ", ".join(chain(
                ("{}={!r}".format(attr, getattr(self, attr))
                for attr in ["id"] + list(self.__class__.__attributes__)
                if hasattr(self, attr)),

                ("{}={}".format(attr, gettimestamp(self, attr))
                for attr in ["created_at", "updated_at"]
                if hasattr(self, attr)))))


    class __metaclass__(type):
        def get_scope(cls, scope_name):
            """
            Retrieve a scope method defined in __scopes__ and set the name
            appropriately.
            """
            # Fetch the scope from the __scopes__ dictionary
            scope = cls.__scopes__[scope_name]
            # Since the scopes defined in __scopes__ are often lambdas
            # to give the name meaning under repr, change the name of the
            # function to "<scope>scope_name"
            scope.__name__ = "<scope>{}".format(scope_name)
            return scope

        @property
        def __all_attributes__(cls):
            attrs = dict(cls.__attributes__)
            attrs.update({
                "created_at": typecasts.datetime,
                "updated_at": typecasts.datetime,
            })
            return attrs

        def __len__(cls):
            return len(Query(cls).all())

        def __getattr__(cls, attr):
            # Is the attr a scope?
            if attr in cls.__scopes__:
                # The attribute is a scope: fetch it, then bind it to the class
                # This way, it is already defined on the class for the next
                # lookup.
                # Think of it like the method_missing + define_method idiom
                # in Ruby (this is the define_method part).
                # "classmethod" handles the binding of the first argument
                setattr(cls, attr, classmethod(cls.get_scope(attr)))
                # Having defined the method, look again: it will be found under
                # normal object lookup
                return getattr(cls, attr)
            elif hasattr(Query(cls), attr):
                return getattr(Query(cls), attr)
            else:
                # The attribute is not a scope: without __getattr__ defined,
                # the behavior would be to raise AttributeError, so that's what
                # we do here. Note that the ususal call to __getattribute__
                # won't work, since it is not define on the metaclass. "super"
                # won't work for the same reason.
                raise AttributeError("'{}' has no attribute '{}'".format(
                    cls.__name__, attr))



import unittest
import mock
import sys
import os
sys.path.insert(0, os.path.dirname(os.path.abspath(os.path.dirname(__file__))))
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.abspath(os.path.dirname(__file__))),
    "lazy_record"))
from lazy_record.associations import *
import lazy_record


class Base(object):
    __dependents__ = []
    __attributes__ = {}
    def __init__(self):
        self._related_records = []


@has_many("comments")
@has_many("test_models", foreign_key="postId")
@has_many("tags", through="taggings")
class Post(Base):

    def boo(self):
        return "blah"

@has_many("things", through="test_models")
@belongs_to("post")
@has_many("test_models")
class Comment(Base):

    def foo(self):
        return "bar"

    @classmethod
    def bar(Comment):
        return "baz"

@has_one("other_thang", through="thing")
@has_one("thing")
@belongs_to("comment")
@belongs_to("post", foreign_key="postId")
class TestModel(Base):
    pass


@belongs_to("post")
@belongs_to("tag")
class Tagging(Base):
    pass


@has_many("another_things")
@has_many("posts", through="taggings")
class Tag(Base):
    pass

@has_one("other_thang", foreign_key="thingId")
@belongs_to("test_model")
class Thing(Base):
    pass

@belongs_to("thing", foreign_key="thingId")
class OtherThang(Base):
    pass

class AnotherThing(Base):
    pass

@has_many("scoped_childs",
          lambda query: query.where(year=2016),
          through="scoping_parents")
class DeepScopingParent(Base):
    pass

@has_many("scoped_childs", lambda query: query.where(name="foo"))
class ScopingParent(Base):
    pass

@belongs_to("scoping_parent")
class ScopedChild(Base):
    pass

@mock.patch("lazy_record.associations.query")
class TestBelongsTo(unittest.TestCase):
    def setUp(self):
        self.comment = Comment()
        self.comment.post_id = 1

    def test_defines_related_method(self, query):
        assert hasattr(self.comment, "post")

    def test_does_not_destroy_other_methods(self, query):
        self.assertEqual("bar", self.comment.foo())

    def test_does_not_destroy_class_methods(self, query):
        self.assertEqual("baz", Comment.bar())

    def test_makes_query_for_related_object(self, query):
        self.comment.post
        query.Query.assert_called_with(Post)
        q = query.Query.return_value
        q.where.assert_called_once_with(id=1)
        q2 = q.where.return_value
        q2.first.assert_called_once_with()

    def test_does_not_change_class_name(self, query):
        self.assertEqual(TestModel.__name__, "TestModel")

    def test_works_with_multiple_belongs_tos(self, query):
        test_model = TestModel()
        test_model.postId = 1
        test_model.comment_id = 1
        assert hasattr(test_model, "post")
        assert hasattr(test_model, "comment")

    def test_multiple_makes_correct_queries(self, query):
        test_model = TestModel()
        test_model.postId = 1
        test_model.comment_id = 17
        q = query.Query.return_value
        q2 = q.where.return_value
        # Test the first
        test_model.post
        query.Query.assert_called_with(Post)
        q.where.assert_called_with(id=1)
        q2.first.assert_called_with()
        # Test the second
        test_model.comment
        query.Query.assert_called_with(Comment)
        q.where.assert_called_with(id=17)
        q2.first.assert_called_with()

    def test_adds_foreign_key_to_attributes(self, query):
        self.assertEqual(TestModel.__attributes__["postId"], int)
        self.assertEqual(Comment.__attributes__["post_id"], int)

    def test_does_not_add_parent_as_dependents(self, query):
        self.assertNotIn("post", Comment.__dependents__)

    def test_allows_changing_of_parent_directly(self, query):
        new_post = Post()
        new_post.id = 87
        self.comment.post = new_post
        self.assertEqual(self.comment.post_id, 87)

    def test_setting_record_with_mismatch_raises(self, query):
        with self.assertRaises(lazy_record.AssociationTypeMismatch):
            not_post = Thing()
            not_post.id = 81
            self.comment.post = not_post

    def test_allows_changing_of_parent_to_None(self, query):
        self.comment.post = None
        self.assertEqual(self.comment.post_id, None)

    def test_adds_entry_to_relationships(self, query):
        self.assertIn("post", associations_for(Comment))


@mock.patch("lazy_record.associations.query")
class TestHasMany(unittest.TestCase):
    def setUp(self):
        self.post = Post()
        self.post.id = 11

    def test_defines_related_method(self, query):
        assert hasattr(self.post, "test_models")

    def test_does_not_destroy_other_methods(self, query):
        self.assertEqual("blah", self.post.boo())

    def test_does_not_destroy_class_methods(self, query):
        self.assertEqual("baz", Comment.bar())

    def test_makes_query_for_related_object(self, query):
        self.post.test_models
        query.Query.assert_called_with(TestModel, record=self.post)
        q = query.Query.return_value
        q.where.assert_called_once_with(postId=11)

    def test_does_not_change_class_name(self, query):
        self.assertEqual(TestModel.__name__, "TestModel")

    def test_works_with_multiple_has_manys(self, query):
        assert hasattr(self.post, "comments")
        assert hasattr(self.post, "test_models")

    def test_multiple_makes_correct_queries(self, query):
        q = query.Query.return_value
        # Test the first
        self.post.comments
        query.Query.assert_called_with(Comment, record=self.post)
        q.where.assert_called_with(post_id=11)
        # Test the second
        self.post.test_models
        query.Query.assert_called_with(TestModel, record=self.post)
        q.where.assert_called_with(postId=11)

    def test_adds_children_as_dependents_when_not_joined(self, query):
        self.assertIn("comments", Post.__dependents__)

    def test_adds_entry_to_relationships(self, query):
        self.assertIn("comments", associations_for(Post))
        self.assertEqual(associations_for(Post)["comments"], None)

    def test_adds_foreign_keys_for_child(self, query):
        self.assertIn("tag", foreign_keys_for(AnotherThing))
        self.assertEqual("tag_id", foreign_keys_for(AnotherThing)['tag'])

    def test_adds_associations_for_child(self, query):
        self.assertIn("tag", associations_for(AnotherThing))
        self.assertEqual(None, associations_for(AnotherThing)['tag'])
        

@mock.patch("lazy_record.associations.query")
class TestHasManyThrough(unittest.TestCase):
    def setUp(self):
        self.post = Post()
        self.post.id = 11

    def test_makes_query_for_related_objects(self, query):
        self.post.tags
        query.Query.assert_called_with(Tag, record=self.post)
        q = query.Query.return_value
        q.joins.assert_called_with("posts")
        q2 = q.joins.return_value
        q2.where.assert_called_once_with(posts=dict(id=11))

    def test_adds_joining_table_as_dependent(self, query):
        self.assertIn("taggings", Post.__dependents__)

    def test_adds_methods_for_joining_table(self, query):
        assert hasattr(self.post, "taggings")

    def test_adds_entry_to_relationships(self, query):
        self.assertIn("tags", associations_for(Post))
        self.assertEqual(associations_for(Post)["tags"], "taggings")        

    def test_adds_association_for_child(self, query):
        self.assertIn("comment", associations_for(Thing))
        self.assertEqual("test_model", associations_for(Thing)['comment'])
        

@mock.patch("lazy_record.associations.query")
class TestHasOne(unittest.TestCase):

    def setUp(self):
        self.test_model = TestModel()
        self.test_model.id = 11
        self.thing = Thing()
        self.thing.id = 17

    def test_makes_query_for_child_object(self, query):
        self.test_model.thing
        query.Query.assert_called_with(Thing, record=self.test_model)
        q = query.Query.return_value
        q.where.assert_called_with(test_model_id=11)

    def test_gets_one_record(self, query):
        q = query.Query.return_value.where.return_value
        record = q.first.return_value
        self.assertEqual(self.test_model.thing, record)
        q.first.assert_called_with()

    def test_adds_child_as_dependent(self, query):
        self.assertIn("thing", TestModel.__dependents__)

    def test_adds_child_to_relationships(self, query):
        self.assertIn("thing", associations_for(TestModel))
        self.assertEqual(associations_for(TestModel)["thing"], None)

    def test_getting_using_custom_foreign_key(self, query):
        self.thing.other_thang
        query.Query.assert_called_with(OtherThang, record=self.thing)
        q = query.Query.return_value
        q.where.assert_called_with(thingId=17)

    def test_setting_child_record(self, query):
        other_thang = OtherThang()
        other_thang2 = OtherThang()
        self.thing.other_thang = other_thang
        self.assertIn(other_thang, self.thing._related_records)
        self.assertEqual(other_thang.thingId, 17)
        query.Query.return_value.where.return_value.first.return_value = \
            other_thang
        self.thing.other_thang = other_thang2
        self.assertIn(other_thang2, self.thing._related_records)
        self.assertEqual(other_thang.thingId, None)
        self.assertEqual(other_thang2.thingId, 17)

    def test_setting_child_record_for_first_time(self, query):
        query.Query.return_value.where.return_value.first.return_value = None
        other_thang = OtherThang()
        self.thing.other_thang = other_thang
        self.assertIn(other_thang, self.thing._related_records)
        self.assertEqual(other_thang.thingId, 17)

    def test_setting_child_raises_if_types_dont_match(self, query):
        with self.assertRaises(lazy_record.AssociationTypeMismatch):
            self.thing.other_thang = Post()

    def test_adds_foreign_key(self, query):
        self.assertIn("other_thang", foreign_keys_for(Thing))
        self.assertEqual("thingId", foreign_keys_for(Thing)["other_thang"])

    def test_adds_foreign_key_to_child(self, query):
        self.assertIn("thing", foreign_keys_for(OtherThang))
        self.assertEqual("thingId", foreign_keys_for(OtherThang)["thing"])


@mock.patch("lazy_record.associations.query")
class TestHasOneThroughOne(unittest.TestCase):

    def test_joins_middle_records(self, query):
        t = TestModel()
        t.id = 93
        t.other_thang
        query.Query.assert_called_with(OtherThang, record=t)
        q = query.Query.return_value
        q.joins.assert_called_with("test_models")
        j = q.joins.return_value
        j.where.assert_called_with(test_models={"id": 93})
        w = j.where.return_value
        w.first.assert_called_with()

    def test_adds_association_to_child(self, query):
        self.assertIn("test_model", associations_for(OtherThang))
        self.assertEqual("thing", associations_for(OtherThang)["test_model"])

@mock.patch("lazy_record.associations.query")
class TestHasManyThroughOne(unittest.TestCase):

    def test_joins_middle_records(self, query):
        comment = Comment()
        comment.id = 16
        comment.things
        query.Query.assert_called_with(Thing, record=comment)
        q = query.Query.return_value
        q.joins.assert_called_with("comments")
        j = q.joins.return_value
        j.where.assert_called_with(comments={"id": 16})

class TestHasOneThroughMany(unittest.TestCase):

    def test_forbids_one_through_many(self):
        
        with self.assertRaises(lazy_record.AssociationForbidden) as e:
            @has_one("thing", through="test_models")
            @has_many("test_models")
            class Forbidden(Base):
                pass

        self.assertEqual(e.exception.message,
                         "Cannot have one 'thing' through many 'test_models'")

@mock.patch("lazy_record.associations.query")
class TestScopedAssociations(unittest.TestCase):

    def test_applies_scope(self, query):
        scoping_parent = ScopingParent()
        scoping_parent.id = 64
        scoping_parent.scoped_childs
        query.Query.assert_called_with(ScopedChild, record=scoping_parent)
        q = query.Query.return_value
        q.where.assert_called_once_with(scoping_parent_id=64)
        w = q.where.return_value
        # Test applies scope
        w.where.assert_called_with(name="foo")

    def test_applies_all_scopes_in_chain(self, query):
        dsp = DeepScopingParent()
        dsp.id = 56
        dsp.scoped_childs
        query.Query.assert_called_with(ScopedChild, record=dsp)
        q = query.Query.return_value
        q.joins.assert_called_with("deep_scoping_parents")
        j = q.joins.return_value
        j.where.assert_called_with(deep_scoping_parents={'id': 56})
        w = j.where.return_value
        # Test that the top level scope is applied
        w.where.assert_called_with(year=2016)
        s = w.where.return_value
        # Test that the through scope is applied
        s.where.assert_called_with(name="foo")

class TestFunctions(unittest.TestCase):

    def test_model_has_foreign_key_for_table_when_has_foreign_key(self):
        self.assertTrue(model_has_foreign_key_for_table("posts", Tagging))

    def test_model_has_foreign_key_for_table_when_not(self):
        self.assertFalse(model_has_foreign_key_for_table("things", TestModel))

if __name__ == '__main__':
    unittest.main()



import unittest
import mock
import sys
import os
sys.path.append(os.path.join(
    os.path.dirname(os.path.abspath(os.path.dirname(__file__))),
    "lazy_record"))
import base
from base import Base
import lazy_record


class MyModel(Base):
    __attributes__ = {
        "name": str,
    }
    __validates__ = {
        "name": lambda record: record.name != "invalid"
    }

    def my_childs():
        pass

class MyOtherModel(Base):
    pass

@mock.patch("base.datetime")
@mock.patch("base.Repo")
@mock.patch("base.Query")
class TestBase(unittest.TestCase):

    def test_gets_number_of_records(self, Query, Repo, datetime):
        Query.return_value.all.return_value.__len__.return_value = 3
        self.assertEqual(len(MyModel), 3)
        Query.assert_called_with(MyModel)
        q = Query.return_value
        q.all.assert_called_once_with()
        all = q.all.return_value
        all.__len__.assert_called_once_with()

    def test_creates_records(self, Query, Repo, datetime):
        Repo.table_name.return_value = "my_model"
        my_record = MyModel(name="me")
        my_record.save()
        Repo.table_name.assert_called_with(MyModel)
        Repo.assert_called_with("my_model")
        repo = Repo.return_value
        today = datetime.datetime.today.return_value
        repo.insert.assert_called_with(name="me",
                                       created_at=today,
                                       updated_at=today)

    def test_updates_records(self, Query, Repo, datetime):
        Repo.table_name.return_value = "my_model"
        my_record = MyModel(name="foo")
        my_record._id = 3
        my_record._created_at = datetime.datetime.today.return_value
        my_record.save()
        Repo.table_name.assert_called_with(MyModel)
        Repo.assert_called_with("my_model")
        repo = Repo.return_value
        repo.where.assert_called_with(id=3)
        where = repo.where.return_value
        today = datetime.datetime.today.return_value
        where.update.assert_called_with(name="foo",
                                        created_at=today,
                                        updated_at=today)

    def test_does_not_create_invalid_records(self, Query, Repo, datetime):
        Repo.table_name.return_value = "my_model"
        my_record = MyModel(name="invalid")
        with self.assertRaises(base.RecordInvalid):
            my_record.save()
        self.assertEqual(Repo.return_value.insert.call_count, 0)

    def test_does_not_update_invalid_records(self, Query, Repo, datetime):
        Repo.table_name.return_value = "my_model"
        my_record = MyModel(name="invalid")
        my_record._id = 3
        my_record._created_at = datetime.datetime.today.return_value
        with self.assertRaises(base.RecordInvalid):
            my_record.save()
        self.assertEqual(Repo.return_value.update.call_count, 0)

    def test_deletes_records(self, Query, Repo, datetime):
        Repo.table_name.return_value = "my_model"
        my_record = MyModel(name="foo")
        my_record._id = 3
        my_record._created_at = datetime.datetime.today.return_value
        my_record.delete()
        Repo.table_name.assert_called_with(MyModel)
        Repo.assert_called_with("my_model")
        repo = Repo.return_value
        repo.where.assert_called_with(id=3)
        where = repo.where.return_value
        where.delete.assert_called_with()

    def test_allows_finding_of_records_by_id(self, Query, Repo, datetime):
        MyModel.find(1)
        Query.assert_called_with(MyModel)
        query = Query.return_value
        query.find.assert_called_with(1)

    def test_allows_finding_of_records_by_attribute(self, Query, Repo, dt):
        MyModel.find_by(name="foo")
        Query.assert_called_with(MyModel)
        query = Query.return_value
        query.find_by.assert_called_with(name="foo")

    def test_allows_searching_of_records_by_attribute(self, Query, Repo, dt):
        MyModel.where(name="foo")
        Query.assert_called_with(MyModel)
        query = Query.return_value
        query.where.assert_called_with(name="foo")

    def test_allows_custom_where(self, Query, Repo, dt):
        MyModel.where("name LIKE ?", "foo")
        Query.assert_called_with(MyModel)
        query = Query.return_value
        query.where.assert_called_with("name LIKE ?", "foo")

    def test_allows_fetching_of_all_records(self, Query, Repo, datetime):
        MyModel.all()
        Query.assert_called_with(MyModel)
        query = Query.return_value
        query.all.assert_called_with()

    def test_allows_fetching_through_joins(self, Query, Repo, datetime):
        MyModel.joins("my_other_models")
        Query.assert_called_with(MyModel)
        query = Query.return_value
        query.joins.assert_called_with("my_other_models")

    def test_casts_attributes_to_correct_type(self, Query, Repo, datetime):
        m = MyModel(name=1)
        self.assertEqual(m.name, "1")

    def test_creates_from_dictionary(self, Query, Repo, datetime):
        m = MyModel.from_dict(id=1, name="foo",
                              created_at=datetime.datetime.today.return_value)
        self.assertEqual(m.id, 1)
        self.assertEqual(m.name, "foo")
        self.assertEqual(m.created_at, datetime.datetime.today.return_value)

    def test_forbids_setting_of_id(self, Query, Repo, datetime):
        m = MyModel()
        with self.assertRaises(AttributeError):
            m.id = 15

    def test_forbids_setting_of_created_at(self, Query, Repo, datetime):
        m = MyModel()
        with self.assertRaises(AttributeError):
            m.created_at = datetime.datetime.today.return_value

    def test_allows_setting_of_attributes(self, Query, Repo, datetime):
        m = MyModel()
        m.name = "turnip"
        self.assertEqual(m.name, "turnip")

    def test_forbits_instantiation_with_id(self, Query, Repo, datetime):
        with self.assertRaises(AttributeError):
            MyModel(id=3)

    def test_forbits_instantiation_with_created_at(self, Query, Repo, dt):
        with self.assertRaises(AttributeError):
            MyModel(created_at=3)

    def test_gets_first_record(self, Query, Repo, datetime):
        MyModel.first()
        Query.assert_called_with(MyModel)
        query = Query.return_value
        query.first.assert_called_with()

    def test_gets_last_record(self, Query, Repo, datetime):
        MyModel.last()
        Query.assert_called_with(MyModel)
        query = Query.return_value
        query.last.assert_called_with()

    def test_mass_assigns_records(self, Query, Repo, datetime):
        m = MyModel()
        m.update(name="foo")
        self.assertEqual(m.name, "foo")

    def test_get_does_not_cast_attr_if_none(self, Query, Repo, datetime):
        m = MyModel()
        self.assertEqual(m.name, None)

    def test_set_does_not_cast_attr_if_none(self, Query, Repo, datetime):
        m = MyModel()
        m.name = None
        self.assertEqual(m.name, None)

    def test_gets_other_attributes_without_cast(self, Query, Repo, datetime):
        m = MyModel()
        self.assertEqual(m.__class__, MyModel)

    def test_raises_if_attribute_not_found(self, Query, Repo, datetime):
        m = MyModel()
        with self.assertRaises(AttributeError):
            m.turkey

    def test_repr_displays_meaningful_represenation(self, Query, Repo, dt):
        m = MyModel()
        self.assertEqual(repr(m),
                         "MyModel(id=None, name=None, created_at=None, "
                         "updated_at=None)")

    def test_evaluates_equality_based_on_id(self, Query, Repo, dt):
        m1 = MyModel()
        m1._id = 1
        m2 = MyModel()
        m2._id = 1
        m3 = MyModel()
        m3._id = 2
        self.assertEqual(m1, m2)
        self.assertNotEqual(m1, m3)

    def test_evaluates_as_inequal_if_either_id_is_None(self, Query, Repo, dt):
        m1 = MyModel()
        m1._id = 1
        m2 = MyModel()
        m2._id = None
        m3 = MyModel()
        m3._id = None
        self.assertNotEqual(m1, m2)
        self.assertNotEqual(m2, m1)
        self.assertNotEqual(m2, m3)

    def test_evaluates_as_inequal_if_models_are_different(self, Q, R, dt):
        m1 = MyModel()
        m1._id = 1
        m2 = MyOtherModel()
        m2._id = 1
        self.assertNotEqual(m1, m2)

    def test_identical_records_evaluate_as_equal(self, Query, Repo, dt):
        m1 = MyModel()
        m1._id = None
        self.assertEqual(m1, m1)

    def test_casts_to_int_as_id(self, Query, Repo, dt):
        m = MyModel()
        m._id = 11
        self.assertEqual(int(m), 11)

    def test_casts_to_0_if_no_id(self, Query, Repo, dt):
        m = MyModel()
        m._id = None
        self.assertEqual(int(m), 0)

    def test_valid_returns_true_when_valid(self, Query, Repo, datetime):
        m = MyModel(name="valid")
        self.assertTrue(m.is_valid())

    def test_valid_returns_false_when_invalid(self, Query, Repo, datetime):
        m = MyModel(name="invalid")
        self.assertFalse(m.is_valid())

    def test_accessing_attribute_not_loaded_raises(self, Query, Repo, dt):
        m = MyModel.from_dict(id=1)
        with self.assertRaises(lazy_record.MissingAttributeError):
            m.name

    def test_accessing_attribute_not_loaded_raises_no_id(self, Query, R, dt):
        m = MyModel.from_dict(name="foo")
        with self.assertRaises(lazy_record.MissingAttributeError) as e:
            m.id
        self.assertEqual(e.exception.message,
                         "'MyModel' object has no attribute 'id'")

    def test_repr_without_timestamps(self, Query, Repo, datetime):
        m = MyModel.from_dict(name="foo")
        self.assertEqual(repr(m), "MyModel(name='foo')")

    def test_create_makes_record(self, Query, Repo, datetime):
        with mock.patch.object(MyModel.__metaclass__, "__call__") as model:
            MyModel.create(name="foo")
            Query(MyModel).create.assert_called_with(name="foo")

    def test_create_returns_record(self, Query, Repo, datetime):
        record = mock.Mock(name="record")
        Query(MyModel).create.return_value = record
        self.assertEqual(MyModel.create(), record)


@mock.patch("base.Repo")
class TestBaseDestroy(unittest.TestCase):

    def setUp(self):
        self.my_model = MyModel(name="hi")
        self.my_model._id = 5

    def test_deletes_without_dependents(self, Repo):
        self.my_model.destroy()
        Repo.assert_called_once_with("my_models")
        repo = Repo.return_value
        repo.where.assert_called_once_with(id=5)
        where = repo.where.return_value
        where.delete.assert_called_once_with()

    @mock.patch.object(MyModel, "__dependents__", new=["my_childs"])
    def test_deletes_dependents(self, Repo):
        my_childs = mock.PropertyMock()
        type(self.my_model).my_childs = my_childs
        child = mock.Mock()
        my_childs.return_value.__iter__.return_value = [child]
        self.my_model.destroy()
        my_childs.assert_called_with()
        child._do_destroy.assert_called_with()
        Repo.assert_called_with("my_models")
        repo = Repo.return_value
        repo.where.assert_called_with(id=5)
        where = repo.where.return_value
        where.delete.assert_called_with()

if __name__ == '__main__':
    unittest.main()



from setuptools import setup
version = "0.6.0"
setup(
  name='lazy_record',
  packages=['lazy_record', 'lazy_record/base'],
  version=version,
  description='Generic Model Layer for Python Web Applications using Sqlite3',
  author='Chase Conklin',
  author_email='cconkli4@u.rochester.edu',
  url='https://github.com/ECESeniorDesign/lazy_record',
  download_url='https://github.com/ECESeniorDesign/'
               'lazy_record/tarball/' + version,
  keywords=['flask'],
  install_requires=[
    'inflector',
  ],
  classifiers=[],
)



import re
import sqlite3
from itertools import chain
from inflector import Inflector, English

inflector = Inflector(English)


class Invalid(Exception):
    pass


class Repo(object):
    """
    Wrapper object around the database.
    """
    db = None

    def __init__(self, table_name):
        """
        Instantiates a Repo object for the passed +table_name+ for adding,
        updating, or destroying records in that table.
        """
        self.table_name = table_name
        self.where_clause = ""
        self.where_values = []
        self.inner_joins = []
        self.order_clause = ""
        self.group_clause = ""
        self.having_clause = ""
        self.having_values = []
        self.limit_value = []

    def where(self, custom_restrictions=[], **restrictions):
        """
        Analog to SQL "WHERE". Does not perform a query until `select` is
        called. Returns a repo object. Options selected through keyword
        arguments are assumed to use == unles the value is a list, tuple, or
        dictionary. List or tuple values translate to an SQL `IN` over those
        values, and a dictionary looks up under a different table when joined.

        ex)

        >>> Repo("foos").where(id=11).select("*")
        SELECT foos.* FROM foos WHERE foos.id == 11
        >>> Repo("foos").where([("id > ?", 12)]).select("*")
        SELECT foos.* FROM foos WHERE foos.id > 12
        >>> Repo("foos").where(id=[1,2,3]).select("*")
        SELECT foos.* FROM foos WHERE foos.id IN (1, 2, 3)
        """
        # Generate the SQL pieces and the relevant values
        standard_names, standard_values = self._standard_items(restrictions)
        custom_names, custom_values = self._custom_items(custom_restrictions)
        in_names, in_values = self._in_items(restrictions)
        query_names = standard_names + custom_names + in_names
        # Stitch them into a clause with values
        if query_names:
            self.where_values = standard_values + custom_values + in_values
            self.where_clause = "where {query} ".format(
                query=" and ".join(query_names))
        return self

    def _in_items(self, restrictions):
        """Generate argument pairs for queries like where(id=[1, 2])"""
        def build_in(table, name, value):
            return "{}.{} IN ({})".format(table, name,
                                          ", ".join(["?"] * len(value)))

        in_items = self._build_where(restrictions, for_in=True)
        names = [build_in(*restriction) for restriction in in_items]
        values = list(chain(*[item[2] for item in in_items]))
        return (names, values)

    def _custom_items(self, restrictions):
        """Generate argument pairs for queries like where("id > ?", 7)"""
        def scope_name(query, table):
            # The first entry in the query is the column
            # If the column already has a ".", that means that the table has
            # already been chosen
            for splitter in (" and ", " or "):
                split_query = re.split(splitter, query, re.IGNORECASE)
                query = splitter.join("{}.{}".format(table, entry)
                                      if "." not in entry else entry
                                      for entry in split_query)
            return query

        names = [scope_name(restriction[0], self.table_name)
                 for restriction in restrictions]
        values = list(chain(
            *[restriction[1:] for restriction in restrictions]))
        return (names, values)

    def _standard_items(self, restrictions):
        """Generate argument pairs for queries like where(id=2)"""
        standard_items = self._build_where(restrictions, for_in=False)
        names = ["{}.{} == ?".format(pair[0], pair[1])
                 for pair in standard_items]
        values = [item[2] for item in standard_items]
        return (names, values)

    def _build_where(self, where_query, for_in):
        # Recursively loops through the where query to produce a list of
        # 3-tuples that contain the (table name, column, value)
        def builder(where_dict, default_table, for_in):
            for key, value in where_dict.items():
                use_in = type(value) in (tuple, list)
                if type(value) is dict:
                    for entry in builder(value, key, for_in):
                        yield entry
                elif (use_in and for_in or not (use_in or for_in)):
                    yield (default_table, key, value)

        return list(builder(where_query, self.table_name, for_in))

    def inner_join(self, *joiners):
        """
        Analog to SQL "INNER JOIN". +joiners+ is a list with entries of the
        form:

        {
            'table': <table_name>,
            'on': [<foreign_key>, <local_id>]
        }

        Example:

        >>> Repo('bs').inner_join(
            {'table': 'cs', on: ['b_id', 'id']}).select("*")
        SELECT bs.* FROM bs INNER JOIN cs ON cs.b_id == bs.id
        """
        def inner_joins(js, current_table):
            for joiner in js:
                yield (((current_table, joiner['on'][1]),
                        (joiner['table'], joiner['on'][0])))
                current_table = joiner['table']

        self.inner_joins = list(inner_joins(joiners, self.table_name))
        return self

    def order_by(self, **kwargs):
        """
        Analog to SQL "ORDER BY". +kwargs+ should only contain one item.

        examples)

        NO:  repo.order_by()
        NO:  repo.order_by(id="desc", name="asc")

        YES: repo.order_by(id="asc)
        """
        if kwargs:
            col, order = kwargs.popitem()
            self.order_clause = "order by {col} {order} ".format(
                col=col, order=order)
        return self

    def group_by(self, column):
        if column:
            self.group_clause = "GROUP BY {} ".format(column)
        return self

    def having(self, conditions):
        names = [condition[0] for condition in conditions]
        self.having_values = list(chain(
            *[condition[1:] for condition in conditions]))
        self.having_clause = "HAVING {query} ".format(
            query=" and ".join(names))
        return self

    def limit(self, count):
        """Limit number of returned rows."""
        if count == 0:
            raise Invalid("Cannot limit to 0 records.")
        self.limit_value = [count]
        return self

    @property
    def join_clause(self):
        # Internal use only, but the API should be stable, except for when we
        # add support for multi-level joins
        return "".join(("inner join {foreign_table} on "
                "{foreign_table}.{foreign_on} == "
                "{local_table}.{local_on} ").format(
                   foreign_table=inner_join[1][0],
                   foreign_on=inner_join[1][1],
                   local_table=inner_join[0][0],
                   local_on=inner_join[0][1],
               ) for inner_join in self.inner_joins)

    @property
    def limit_clause(self):
        if self.limit_value != []:
            return "LIMIT ? "
        else:
            return ""

    def select(self, *attributes):
        """
        Select the passed +attributes+ from the table, subject to the
        restrictions provided by the other methods in this class.

        ex)

        >>> Repo("foos").select("name", "id")
        SELECT foos.name, foos.id FROM foos
        """
        namespaced_attributes = [
            "{table}.{attr}".format(table=self.table_name, attr=attr)
            for attr in attributes
        ]
        cmd = ('select {attrs} from {table} '
               '{join_clause}{where_clause}{order_clause}'
               '{group_clause}{having_clause}{limit_clause}').format(
            table=self.table_name,
            attrs=", ".join(namespaced_attributes),
            where_clause=self.where_clause,
            join_clause=self.join_clause,
            order_clause=self.order_clause,
            group_clause=self.group_clause,
            having_clause=self.having_clause,
            limit_clause=self.limit_clause,
        ).rstrip()
        return Repo.db.execute(cmd, self.where_values + self.having_values + \
                               self.limit_value)

    def count(self):
        """
        Count the number of records in the table, subject to the query.
        """
        cmd = ("select COUNT(*) from {table} "
               "{join_clause}{where_clause}{order_clause}").format(
                    table=self.table_name,
                    where_clause=self.where_clause,
                    join_clause=self.join_clause,
                    order_clause=self.order_clause).rstrip()
        return Repo.db.execute(cmd, self.where_values)

    def insert(self, **data):
        """
        Insert the passed +data+ into the table. Raises Invalid if a where
        clause is present (i.e. no INSERT INTO table WHERE)
        """
        if self.where_clause:
            raise Invalid("Cannot insert with 'where' clause.")
        # Ensure that order is preserved
        data = data.items()
        cmd = "insert into {table} ({attrs}) values ({values})".format(
            table=self.table_name,
            attrs=", ".join(entry[0] for entry in data),
            values=", ".join(["?"] * len(data)),
        )
        handle = Repo.db.execute(cmd, [entry[1] for entry in data])
        # Return the id of the added row
        return handle.lastrowid

    def update(self, **data):
        """
        Update records in the table with +data+. Often combined with `where`,
        as it acts on all records in the table unless restricted.

        ex)

        >>> Repo("foos").update(name="bar")
        UPDATE foos SET name = "bar"
        """
        data = data.items()
        update_command_arg = ", ".join("{} = ?".format(entry[0])
                                       for entry in data)
        cmd = "update {table} set {update_command_arg} {where_clause}".format(
            update_command_arg=update_command_arg,
            where_clause=self.where_clause,
            table=self.table_name).rstrip()
        Repo.db.execute(cmd, [entry[1] for entry in data] + self.where_values)

    def delete(self):
        """
        Remove entries from the table. Often combined with `where`, as it acts
        on all records in the table unless restricted.
        """
        cmd = "delete from {table} {where_clause}".format(
            table=self.table_name,
            where_clause=self.where_clause
        ).rstrip()
        Repo.db.execute(cmd, self.where_values)

    @staticmethod
    def table_name(model):
        """
        Get a model's table name. (e.g. MyModel => "my_models")
        """
        return inflector.tableize(model.__name__)

    @classmethod
    def connect_db(Repo, database=":memory:"):
        """
        Connect Repo to a database with path +database+ so all instances can
        interact with the database.
        """
        Repo.db = sqlite3.connect(database,
                                  detect_types=sqlite3.PARSE_DECLTYPES)
        return Repo.db
